import jsonpickle

# 需要转换成str的属性
str_value_keys = []
jsonpickle.set_preferred_backend('json')
jsonpickle.set_encoder_options('json', ensure_ascii=False)
jsonpickle.set_decoder_options('json')

def dumps(obj, **kwargs) -> str:
  """
  将对象转换为json字符串
  :param obj: 对象
  :return: json 字符串
  """
  if obj is None: return None
  if type(obj) == str: return obj
  return f'{jsonpickle.encode(obj, unpicklable=False, make_refs=False, **kwargs)}'

def loads(json_str: str, **kwargs) -> dict:
  """
  将json字符串转换为对象
  :param json_str: json 字符串
  :return: 对象
  """
  return jsonpickle.decode(json_str, **kwargs)

def unify_to_str(json_str: str) -> str:
  if not str_value_keys and len(str_value_keys) == 0: return json_str
  obj = loads(json_str)
  if isinstance(obj, list):
    _handle_list(obj)
  elif isinstance(obj, dict):
    _handle_dict(obj)
  return dumps(obj)

def _handle_list(data):
  for o in data:
    if isinstance(o, list):
      _handle_list(o)
    elif isinstance(o, dict):
      _handle_dict(o)

def _handle_dict(data):
  for k, v in data.items():
    if isinstance(v, list):
      _handle_list(v)
    elif isinstance(v, dict):
      _handle_dict(v)
    elif k in str_value_keys:
      try:
        data[k] = str(v)
      except Exception:
        pass



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2024/9/19 14:02'

import math

from jsonpath_ng import parser

from ctools import cjson
from ctools.sys_log import flog as log

"""
douglas_rarefy.DouglasRarefy(res, level=3).sparse_points()
"""

class THIN_LEVEL:
  L1 = 0.00001
  L2 = 0.00003
  L3 = 0.00009
  L4 = 0.0002
  L5 = 0.0004
  L6 = 0.0007
  L7 = 0.0011
  L8 = 0.0017
  L9 = 0.0022

class Point:
  def __init__(self, lng, lat, origin_data):
    self.lng = lng
    self.lat = lat
    self.origin_data = origin_data

def _get_line_by_point(xy1, xy2):
  """
  根据两个点求直线方程 ax + by + c = 0
  :param xy1: 点1, 例如 {'lat': 1, 'lng': 1}
  :param xy2: 点2, 例如 {'lat': 2, 'lng': 2}
  :return: 直线方程的三个参数 [a, b, c]
  """
  x1 = xy1.lng
  y1 = xy1.lat
  x2 = xy2.lng
  y2 = xy2.lat
  a = y2 - y1
  b = x1 - x2
  c = (y1 - y2) * x1 - y1 * (x1 - x2)
  return [a, b, c]


def _get_distance_from_point_to_line(a, b, c, xy):
  """
  点到直线的距离，直线方程为 ax + by + c = 0
  :param a: 直线参数a
  :param b: 直线参数b
  :param c: 直线参数c
  :param xy: 点坐标，例如 {'lat': 2, 'lng': 2}
  :return: 距离
  """
  x = xy.lng
  y = xy.lat
  return abs((a * x + b * y + c) / math.sqrt(a * a + b * b))


class DouglasRarefy:
  """
  DouglasRarefy Use Guide:
    points must be arrays, element can be dict or arrays, when element is arrays, index 0 must be lng, index 1 must be lat, and element can be use max column num is 2 (lng, lat) in ret_tpl
    level default is L2, this level can be hold most of the points detail
    ret_tpl is the result tpl, can be arrays and json or some can be json loads, exp: [{lng}, {lat}] OR {{"lng": {lng}, "lat": {lat}}}
  """
  def __init__(self, points:[], level=THIN_LEVEL.L2, ret_tpl=None, get_lng=None, get_lat=None):
    if not isinstance(points, list): raise Exception('points must be list obj !!')
    if len(points) < 3: raise Exception('points length must be gt 2 !!')
    self.points = points
    self.threshold = THIN_LEVEL.L2 if level is None else (getattr(THIN_LEVEL, "L{}".format(int(level))) if int(level) >= 1 else level)
    log.debug("threshold is: {}".format(self.threshold))
    self.is_json = isinstance(points[0], dict)
    self.get_lng = get_lng
    self.get_lat = get_lat
    if self.is_json:
      if not self.get_lng: self.get_lng = '$.lng'
      if not self.get_lat: self.get_lat = '$.lat'
    else:
      if not self.get_lng: self.get_lng = '$.[0]'
      if not self.get_lat: self.get_lat = '$.[1]'
    log.debug("get_lng is: {}, get_lat is: {}".format(self.get_lng, self.get_lat))
    self.lng_parser = parser.parse(self.get_lng)
    self.lat_parser = parser.parse(self.get_lat)
    log.debug("is_json is: {}".format(self.is_json))
    self.ret_tpl = ret_tpl
    log.debug("ret_tpl is: {}".format(self.ret_tpl))
    self.data = [Point(self.lng_parser.find(p)[0].value, self.lat_parser.find(p)[0].value, p) for p in self.points]

  def _sparse_points(self, points):
    """
    点位压缩
    :return: 稀疏后的点集
    """
    if len(points) < 3:
      if not self.ret_tpl:
        return [points[0].origin_data, points[-1].origin_data]
      else:
        if self.is_json:
          return [cjson.loads(self.ret_tpl.format(**points[0].origin_data)), cjson.loads(self.ret_tpl.format(**points[-1].origin_data))]
        else:
          return [cjson.loads(self.ret_tpl.format(lng=points[0].lng, lat=points[0].lat)), cjson.loads(self.ret_tpl.format(lng=points[-1].lng, lat=points[-1].lat))]

    xy_first = points[0]   # 第一个点
    xy_end   = points[-1]  # 最后一个点
    a, b, c  = _get_line_by_point(xy_first, xy_end)  # 获取直线方程的 a, b, c 值
    d_max = 0   # 记录点到直线的最大距离
    split = 0   # 分割位置
    for i in range(1, len(points) - 1):
      d = _get_distance_from_point_to_line(a, b, c, points[i])
      if d > d_max:
        split = i
        d_max = d
    if d_max > self.threshold:
      # 如果存在点到首位点连成直线的距离大于 max_distance 的, 即需要再次划分
      child_left  = self._sparse_points(points[:split + 1])  # 递归处理左边部分
      child_right = self._sparse_points(points[split:])      # 递归处理右边部分
      # 合并结果，避免重复
      return child_left + child_right[1:]
    else:
      if not self.ret_tpl:
        return [points[0].origin_data, points[-1].origin_data]
      else:
        if self.is_json:
          return [cjson.loads(self.ret_tpl.format(**points[0].origin_data)), cjson.loads(self.ret_tpl.format(**points[-1].origin_data))]
        else:
          return [cjson.loads(self.ret_tpl.format(lng=points[0].lng, lat=points[0].lat)), cjson.loads(self.ret_tpl.format(lng=points[-1].lng, lat=points[-1].lat))]

  def sparse_points(self):
    return self._sparse_points(self.data)




import tkinter as tk
from multiprocessing import Process, Queue

data_queue: Queue = None
canvas_process: Process = None

class MSGType:
  BOTTOM = 'bottom'       #置于底部
  DRAW = 'draw'           #绘制矩形

class CanvasInfo:
  def __init__(self,title: str='', points: list=None, msg_type: MSGType=MSGType.DRAW):
    self.msg_type = msg_type
    # 标题
    self.title = title
    # 左上, 右上, 右下, 左下
    # [(100, 100), (200, 100), (200, 150), (100, 150)]
    self.points = points

def on_ctrl_click(event):
  print("Ctrl+Left Click detected")

class RecorderTool(tk.Tk):
  def __init__(self, data: Queue):
    super().__init__()
    self.data = data
    self.title('Recorder Tool')
    self.attributes('-topmost', True)
    self.overrideredirect(True)
    self.attributes('-transparentcolor', 'grey')
    self.geometry("{0}x{1}+0+0".format(self.winfo_screenwidth(), self.winfo_screenheight()))
    self.canvas = tk.Canvas(self, highlightthickness=0, bg='grey')
    self.canvas.pack(fill="both", expand=True)
    self.bind('<Control-Button-1>', on_ctrl_click)

    self.check_queue()

  def draw_red_border(self, canvas_info: CanvasInfo):
    self.clear_canvas()
    points = canvas_info.points
    x1, y1 = points[0]
    x2, y2 = points[1]
    x3, y3 = points[2]
    x4, y4 = points[3]
    self.canvas.create_line(x1, y1, x2, y2, fill="red", width=2)
    self.canvas.create_line(x2, y2, x3, y3, fill="red", width=2)
    self.canvas.create_line(x3, y3, x4, y4, fill="red", width=2)
    self.canvas.create_line(x4, y4, x1, y1, fill="red", width=2)
    text_x = (x1 + x2) / 2
    text_y = y1 - 15
    if text_y < 0: text_y = y3 + 15
    self.canvas.create_text(text_x, text_y, text=canvas_info.title, font=("Arial", 12), fill="red")

  def clear_canvas(self):
    self.canvas.delete("all")

  def check_queue(self):
    if not self.data.empty():
      canvas_info = self.data.get_nowait()
      if canvas_info.msg_type == MSGType.BOTTOM:
        self.clear_canvas()
        self.withdraw()
      elif canvas_info.msg_type == MSGType.DRAW:
        self.deiconify()
        self.draw_red_border(canvas_info)
    self.after(10, self.check_queue)

def _init_recorder(data):
  canvas_app = RecorderTool(data)
  canvas_app.mainloop()

def start():
  global data_queue, canvas_process
  data_queue = Queue()
  canvas_process = Process(target=_init_recorder, args=(data_queue,))
  canvas_process.start()
  return canvas_process

def stop():
  canvas_process.terminate()
  canvas_process.join()
  data_queue.close()




import os
from urllib.parse import urlencode

from bottle import static_file, HTTPResponse

from ctools import sys_log, http_utils

log = sys_log.flog

"""
文件下载服务
"""


def download(file_path: str, download_name:str=None):
  """
  文件下载
  :param file_path: 静态文件路径
  :param download_name: 下载文件名
  :return:
  """
  if os.path.exists(file_path):
    root_path = os.path.split(file_path)[0]
    file_name = os.path.split(file_path)[1]
    download_filename = urlencode({'filename': download_name or file_name}).split("=")[-1]  # 对文件名进行URL编码
    response = static_file(file_name, root=root_path, download=True)
    # 设置响应头，告诉浏览器这是一个文件下载
    response.headers['Content-Type'] = 'application/octet-stream;charset=utf-8'
    response.headers['Content-Disposition'] = f'attachment; filename={download_filename}'
    log.debug(f"下载文件成功, file_path： {file_path}, file_name： {file_name}, download_name： {download_name}")
  else:
    response = None
    log.info("下载文件失败, 此文件不存在, file_path： %s" % file_path)
  return response


def download_bytes(file_bytes: bytes, download_name: str):
  """
  文件下载
  :param file_bytes: file_bytes
  :param download_name: download_name
  :return:
  """
  download_filename = urlencode({'filename': download_name}).split("=")[-1]  # 对文件名进行URL编码
  # 设置响应头，告诉浏览器这是一个文件下载
  headers = {"Accept-Ranges": "bytes", "Content-Length": len(file_bytes),
             'Content-Type': 'application/octet-stream;charset=utf-8',
             'Content-Disposition': f'attachment; filename={download_filename}'}
  return HTTPResponse(file_bytes, **headers)


def download_url(url: str, save_path: str):
  content = http_utils.get(url)
  if content:
    with open(save_path, "wb") as f:
      f.write(content)
  return save_path



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2025/1/24 08:53'

import io
import time
from ftplib import FTP, error_perm, error_temp

class CFTP:
  """
  with open('xx/xx.md', 'rb') as file:
    ftp_host = 'x.x.x.x'
    ftp_username = 'x'
    ftp_password = 'x'
    CFTP(ftp_host, ftp_username, ftp_password).upload_file_to_ftp('xx.md', file)
  """
  def __init__(self, host, username, password, timeout=30, max_retries=3, retry_delay=5):
    self.host = host
    self.username = username
    self.password = password
    self.ftp:FTP = None
    self.timeout = timeout
    self.max_retries = max_retries
    self.retry_delay = retry_delay

  def upload_file_to_ftp(self, file_name, file:io.BytesIO, ftp_directory='/'):
    if not file_name: raise Exception('文件名不能为空')
    if not file: raise Exception('文件不能为空')
    for attempt in range(self.max_retries):
      try:
        if not self.ftp:
          ftp = FTP(self.host, timeout=self.timeout)
          ftp.login(self.username, self.password)
          ftp.set_pasv(True)
          self.ftp = ftp
        try:
          self.ftp.cwd(ftp_directory)
        except error_perm:
          print(f"FTP 目录不存在：{ftp_directory}")
          self.ftp.mkd(ftp_directory)
          print(f"FTP 目录创建成功：{ftp_directory}, 正在切换到目录：{ftp_directory}")
          self.ftp.cwd(ftp_directory)
        print(f"正在上传文件：{file_name}")
        self.ftp.storbinary(f"STOR {file_name}", file)
        self.ftp.quit()
        print(f"文件成功上传到 FTP: {file_name}")
        return
      except (error_perm, error_temp) as e:
        try:
          self.ftp.quit()
        except Exception:
          pass
        self.ftp = None
        print(f"FTP 错误：{e}")
        if attempt < self.max_retries - 1:
          print(f"正在尝试重连... 第 {attempt + 1} 次重试...")
          time.sleep(self.retry_delay)
        else:
          print("重试次数已用尽，上传失败。")
          raise
      except Exception as e:
        try:
          self.ftp.quit()
        except Exception:
          pass
        self.ftp = None
        print(f"连接或上传出现异常：{e}")
        if attempt < self.max_retries - 1:
          print(f"正在尝试重连... 第 {attempt + 1} 次重试...")
          time.sleep(self.retry_delay)
        else:
          print("重试次数已用尽，上传失败。")
          raise



#!/usr/bin/env python3.6
# coding: utf-8
import sched
import threading
import time
import traceback
from datetime import datetime
from functools import wraps
from typing import Optional, Dict

import pytz
from croniter import croniter

"""
@cron_lite.cron_task('0/1 * * * * ? *')
def demo():
  print('hello world')

@cron_lite.cron_task('0/1 * * * * ? *')
def demo1():
  print('hello world111')

def demo2(xx, fff):
  print('hello world222', xx, fff)

cron_lite.apply_cron_task('0/1 * * * * ? *', demo2, (123123123, 34534534))
print(123123)

cron_lite.start_all()
"""

class SchedulerMeta:
  timer_task_name: str = None
  switch: bool = True
  status: bool = False
  event: sched.Event = None
  scheduler: sched.scheduler = None


scheduler_map: Dict[str, SchedulerMeta] = {}  # {timer_task_name: SchedulerMeta}
_switch = False
_info_handler  = print
_error_handler = print
_time_zone: Optional[pytz.BaseTzInfo] = None


def set_time_zone(time_zone_name: str):
  global _time_zone
  _time_zone = pytz.timezone(time_zone_name)

# @annotation
def cron_task(cron_expr: str, task_name: str = None, till_time_stamp: int = None):
  """
  cron_task decorator to register a function as crontab task
  :param cron_expr: the croniter accepted cron_expression. NOTICE: the default timezone is UTC and can be changed by
  `set_time_zone`. The format is `min hour day month weekday [sec]`
  :param till_time_stamp: run this jog till when. None means forever
  :return: the real decorator
  """
  cron_expr = _convert_cron(cron_expr)
  assert len(cron_expr.split(" ")) in (5, 6), \
    "only supported <min hour day month weekday> and <min hour day month weekday sec>"

  def deco(func):
    @wraps(func)
    def inner():
      try:
        func()
      except Exception:
        try:
          _error_handler(f"run {func.__name__} failed\n" + traceback.format_exc())
        except Exception:
          _error_handler(f"run {func.__name__} failed\n")
      _register_next(inner.__name__ if task_name is None else task_name, inner, cron_expr, till_time_stamp)

    _register_next(inner.__name__ if task_name is None else task_name, inner, cron_expr, till_time_stamp, init=True)
    return inner

  return deco


def apply_cron_task(cron_expr, func, params, timer_task_name=None, till_time_stamp=None):
  """
  cron_task decorator to register a function as crontab task
  :param func: task callback function
  :param params: transparent parameters
  :param cron_expr: the croniter accepted cron_expression. NOTICE: the default timezone is UTC and can be changed by
  `set_time_zone`. The format is `min hour day month weekday [sec]`
  :param timer_task_name: task name
  :param till_time_stamp: run this jog till when. None means forever
  :return: the real decorator
  """
  cron_expr = _convert_cron(cron_expr)
  assert len(cron_expr.split(" ")) in (5, 6),  "Only supported <minute hour day month weekday> and <minute hour day month weekday second>"
  task_name = func.__name__ if timer_task_name is None else timer_task_name
  @wraps(func)
  def wrapper(*args, **kwargs):
    try:
      nonlocal params
      func.__taskName__ = task_name
      func(*params, *args, **kwargs)
    except Exception as exc:
      _error_handler(f"Run {func.__name__} failed with error: {str(exc)}")
    finally:
      _register_next(task_name, wrapper, cron_expr, till_time_stamp)

  _register_next(task_name, wrapper, cron_expr, till_time_stamp, init=True)
  # 不使用 submit, 因为提交的任务, 不是 daemon 线程
  t = threading.Thread(target=_start, args=(timer_task_name, ))
  t.setDaemon(True)
  t.start()
  return t

def start_all(spawn: bool = True, daemon: bool = True, info_handler=None, error_handler=None) -> Optional[threading.Thread]:
  """
  start_all starts all cron tasks registered before.
  :param spawn: whether to start a new thread for scheduler. If not, the action will block the current thread
  :param daemon: the new thread is daemon if True
  :param info_handler: handle info output (scheduler start / stop), default = print, can use logging.info
  :param error_handler: handle error output (task execute exception), default = print, can use logging.error
  :raise RuntimeError: if the tasks are already started and still running we cannot start again. The feature is not
  concurrent-safe
  :return: the new thread if spawn = True
  """
  global _switch, _info_handler, _error_handler
  if _switch:
    raise RuntimeError("the crontab was already started...")
  if info_handler:
    _info_handler = info_handler
  if error_handler:
    _error_handler = error_handler
  if spawn:
    t = threading.Thread(target=_start)
    t.setDaemon(daemon)
    t.start()
    return t
  else:
    _start()


def is_active(timer_task_name):
  res = False
  if timer_task_name in scheduler_map:
    res = scheduler_map.get(timer_task_name).switch or scheduler_map.get(timer_task_name).status
  return res


def active(timer_task_name):
  if timer_task_name in scheduler_map:
    scheduler_map.get(timer_task_name).status = True

def get_switch(timer_task_name):
  switch = True
  if timer_task_name in scheduler_map:
    switch = scheduler_map.get(timer_task_name).switch
  return switch


def inactive(timer_task_name):
  if timer_task_name in scheduler_map:
    scheduler_map.get(timer_task_name).status = False
    if not scheduler_map.get(timer_task_name).switch:
      scheduler_map.get(timer_task_name).scheduler.cancel(scheduler_map[timer_task_name].event)


def stop(timer_task_name):
  if timer_task_name in scheduler_map:
    scheduler_map.get(timer_task_name).switch = False
    time.sleep(1)


def stop_all(wait_thread: Optional[threading.Thread] = None):
  """
  stop_all turns off the switch to stop the scheduler. Running jobs will be wait till finished.
  :param wait_thread: join() the spawned scheduler thread (if you started it as spawn and you want) to ensure all jobs
  to finish
  :return:
  """
  for timer_task_name in scheduler_map:
    scheduler_map.get(timer_task_name).switch = False
  if wait_thread:
    wait_thread.join()


def _register_next(timer_task_name, base_func, cron_expr, till_time_stamp, init: bool = False):
  cron_obj = croniter(cron_expr)
  if _time_zone:
    cron_obj.set_current(datetime.now(tz=_time_zone))
  next_time = int(cron_obj.get_next())
  if scheduler_map.get(timer_task_name) is None:
    scheduler_meta = SchedulerMeta()
    scheduler_meta.timer_task_name = timer_task_name
    scheduler_meta.switch = True
    scheduler_meta.scheduler = sched.scheduler(time.time, time.sleep)
    scheduler_map[timer_task_name] = scheduler_meta
  elif init:
    raise ValueError(f"task name: {timer_task_name} already exists!!!!!")
  if till_time_stamp is None or next_time <= till_time_stamp:
    scheduler_map[timer_task_name].event = scheduler_map[timer_task_name].scheduler.enterabs(next_time, 0, base_func)


def _run_sched(scheduler_meta: SchedulerMeta):
  active(scheduler_meta.timer_task_name)
  while True:
    scheduler = scheduler_meta.scheduler
    if not _switch or not scheduler_meta.switch:
      scheduler.empty()
      inactive(scheduler_meta.timer_task_name)
      return
    t = scheduler.run(False)
    if t is None:
      inactive(scheduler_meta.timer_task_name)
      return
    st = time.time()
    while time.time() - st < t:
      if not _switch or not scheduler_meta.switch:
        scheduler.empty()
        inactive(scheduler_meta.timer_task_name)
        return
      time.sleep(0.5)


def _start(taskName: str = None):
  global _switch
  _switch = True
  _info_handler("cron job begin start...")
  taskList = []
  for timer_task_name, scheduler_meta in scheduler_map.items():
    if taskName is not None and timer_task_name != taskName: continue
    print("register job: ", timer_task_name)
    thread = threading.Thread(target=_run_sched, args=(scheduler_meta, ))
    thread.setDaemon(True)
    thread.start()
    taskList.append(thread)
  for task in taskList: task.join()
  _info_handler("cron job execute finished...")
  _switch = False
  scheduler_map.clear()


def _convert_cron(cron_expr):
  res_cron = ""
  cron_list = cron_expr.split(" ")
  if len(cron_list) > 6:
    for cron in cron_list[1:]:
      if cron != "?":
        res_cron += "%s " % cron
    res_cron += "%s" % cron_list[0]
  else:
    res_cron = cron_expr
  return res_cron




import hashlib
import hmac

global_key = 'gomyck2014'


def generate_signature(file_path, key: str = global_key):
  try:
    with open(file_path, 'rb') as f:
      file_contents = f.read()
      file_hash = hashlib.sha256(file_contents).digest()
    sign_val = hmac.new(key.encode(), file_hash, hashlib.sha256).digest()
    return sign_val.hex()
  except:
    return ''

def digest(value: str, key: str = global_key):
  val_hash = hashlib.sha256(value.encode()).digest()
  sign_val = hmac.new(key.encode(), val_hash, hashlib.sha256).digest()
  return sign_val.hex()



import time

# 64位ID的划分
WORKER_ID_BITS = 5
DATACENTER_ID_BITS = 5
SEQUENCE_BITS = 12
# 最大取值计算
MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**5-1 0b11111
MAX_DATACENTER_ID = -1 ^ (-1 << DATACENTER_ID_BITS)
# 移位偏移计算
WOKER_ID_SHIFT = SEQUENCE_BITS
DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS
# 序号循环掩码
SEQUENCE_MASK = -1 ^ (-1 << SEQUENCE_BITS)
# Twitter元年时间戳
TWEPOCH = 1288834974657

class SnowId(object):
  """
  用于生成IDs
  """
  def __init__(self, datacenter_id=0, worker_id=0, sequence=0):
    """
    初始化
    :param datacenter_id: 数据中心（机器区域）ID
    :param worker_id: 机器ID
    :param sequence: 起始序号
    """
    # sanity check
    if worker_id > MAX_WORKER_ID or worker_id < 0:
      raise ValueError('worker_id值越界')
    if datacenter_id > MAX_DATACENTER_ID or datacenter_id < 0:
      raise ValueError('datacenter_id值越界')
    self.worker_id = worker_id
    self.datacenter_id = datacenter_id
    self.sequence = sequence
    self.last_timestamp = -1  # 上次计算的时间戳

  @staticmethod
  def _gen_timestamp():
    """
    生成整数时间戳
    :return:int timestamp
    """
    return int(time.time() * 1000)

  def get_id(self):
    """
    获取新ID
    :return:
    """
    timestamp = self._gen_timestamp()
    # 时钟回拨
    if timestamp < self.last_timestamp:
      print('clock is moving backwards. Rejecting requests until {}'.format(self.last_timestamp))
      raise
    if timestamp == self.last_timestamp:
      self.sequence = (self.sequence + 1) & SEQUENCE_MASK
      if self.sequence == 0:
        timestamp = self._til_next_millis(self.last_timestamp)
    else:
      self.sequence = 0
    self.last_timestamp = timestamp
    new_id = ((timestamp - TWEPOCH) << TIMESTAMP_LEFT_SHIFT) | (self.datacenter_id << DATACENTER_ID_SHIFT) | \
             (self.worker_id << WOKER_ID_SHIFT) | self.sequence
    return new_id

  def _til_next_millis(self, last_timestamp):
    """
    等到下一毫秒
    """
    timestamp = self._gen_timestamp()
    while timestamp <= last_timestamp:
      timestamp = self._gen_timestamp()
    return timestamp



import os

from ctools import sys_log, date_utils

log = sys_log.flog

"""
文件上传服务
"""


def save(upload, output_dir: str, file_name: str = None):
  """
  根据上传的upload对象保存上传文件
  :param output_dir:
  :param upload:
  :param file_name:
  :return: 保存文件路径
  """

  # 确保文件夹存在
  os.makedirs(output_dir, exist_ok=True)

  unique_filename = file_name
  if not unique_filename:
    # 生成不重复的文件名，加上时间戳
    file_name, ext = os.path.splitext(upload.raw_filename)
    timestamp = date_utils.get_time()
    unique_filename = f'{file_name}_{timestamp}{ext}'
  else:
    dot_index = unique_filename.find(".")
    suffix = os.path.splitext(upload.raw_filename)[-1]
    if dot_index != -1:
      unique_filename = os.path.splitext(unique_filename)[0] + suffix
    else:
      unique_filename = unique_filename + suffix
  save_path = os.path.join(output_dir, unique_filename)
  upload.save(save_path, overwrite=True)
  log.info("上传文件成功: %s" % save_path)
  return file_name, save_path



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2024/10/25 09:42'

class DictWrapper(dict):
  def __getattr__(self, key):
    res = self.get(key)
    if res is None:
      res = self.query_params.get(key)
      if res is None:
        raise AttributeError(f" ==>> {key} <<== Not Found In This Entity!!!")
    if isinstance(res, dict):
      return DictWrapper(res)
    return res

  def __setattr__(self, key, value):
    self[key] = value

  def __delattr__(self, key):
    del self[key]



import os
import time
import tkinter as tk

import pyautogui

from ctools import application, string_tools

"""
截屏工具类, 按回车截图
image_path=screenshot_tools.screenshot()
"""


class ScreenshotTools:
  def __init__(self):
    self.root = tk.Tk()
    self.root.overrideredirect(True)
    self.root.attributes("-alpha", 0.1)
    self.root.attributes('-topmost', 'true')
    self.root.geometry("{0}x{1}+0+0".format(self.root.winfo_screenwidth(), self.root.winfo_screenheight()))
    self.root.configure(bg="black")

    self.canvas = tk.Canvas(self.root)
    self.canvas.bind("<Button-3>", self.change_center_point)

    self.x, self.y = 0, 0
    self.xstart, self.ystart = 0, 0
    self.xend, self.yend = 0, 0
    self.xcenter, self.ycenter = 0, 0

    self.select_area = self.canvas.create_rectangle(0, 0, 0, 0, outline='red', width=0, dash=(4, 4))
    self.screenshot_path = None
    self.center_offset = (0, 0)

    self.pos_div = tk.Toplevel(self.root)
    self.pos_div.overrideredirect(True)
    self.pos_div.attributes('-topmost', 'true')
    self.pos_div.configure(bg="grey")
    self.pos_div.geometry("360x25")

    self.pos_div_text = "POS: (%s, %s), Enter键-截图, Esc键-退出, 右键设置点击位置"
    self.pos_label = tk.Label(self.pos_div, text=self.pos_div_text % (0, 0), bg="black", fg="white")
    self.pos_label.pack(fill="both", expand=True)

    self.center_div = None

    self.root.bind('<Escape>', self.sys_out)  # 键盘Esc键->退出
    self.root.bind('<Motion>', self.print_pos)
    self.root.bind("<Button-1>", self.mouse_left_down)  # 鼠标左键点击->显示子窗口
    self.root.bind("<B1-Motion>", self.mouse_left_move)  # 鼠标左键移动->改变子窗口大小
    self.root.bind("<ButtonRelease-1>", self.mouse_left_up)  # 鼠标左键释放->记录最后光标的位置
    self.root.bind("<Return>", self.save_image)  # 回车键->截屏并保存图片
    self.root.focus()

  def start(self):
    self.root.mainloop()

  def print_pos(self, event):
    self.x, self.y = event.widget.winfo_pointerxy()
    self.pos_label.config(text=self.pos_div_text % (self.x, self.y))
    self.pos_div.geometry(f"+{self.x + 10}+{self.y + 10}")

  def change_center_point(self, event):
    offset_x = event.x + self.xstart - self.xcenter - 2
    offset_y = event.y + self.ystart - self.ycenter - 2
    self.center_div.geometry(f"+{self.xstart+event.x-2}+{self.ystart+event.y-2}")
    self.center_offset = (offset_x, offset_y)

  def mouse_left_down(self, event):
    if self.center_div is None:
      self.center_div = tk.Toplevel(self.root)
      self.center_div.overrideredirect(True)
      self.center_div.attributes('-topmost', 'true')
      self.center_div.configure(bg="red")
      self.center_div.geometry("4x4")

    self.x, self.y = event.x, event.y
    self.xstart, self.ystart = event.x, event.y
    self.canvas.configure(height=1)
    self.canvas.configure(width=1)
    self.canvas.config(highlightthickness=0)  # 无边框
    self.canvas.place(x=event.x, y=event.y)
    self.center_div.geometry(f"+{self.xstart}+{self.ystart}")

  def mouse_left_move(self, event):
    self.x, self.y = event.x, event.y
    self.xcenter = self.xstart + int((self.x-self.xstart)/2)-2
    self.ycenter = self.ystart + int((self.y-self.ystart)/2)-2

    self.pos_label.config(text=self.pos_div_text % (self.x, self.y))
    self.pos_div.geometry(f"+{self.x + 10}+{self.y + 10}")
    self.center_div.geometry(f"+{self.xcenter}+{self.ycenter}")

    self.canvas.configure(height=event.y - self.ystart)
    self.canvas.configure(width=event.x - self.xstart)
    self.canvas.coords(self.select_area, 0, 0, event.x - self.xstart, event.y - self.ystart)

  def mouse_left_up(self, event):
    self.xend, self.yend = event.x, event.y

  def save_image(self, event):
    try:
      self.canvas.delete(self.select_area)
      self.canvas.place_forget()
      self.sys_out()

      time.sleep(0.3)
      img = pyautogui.screenshot(region=[self.xstart, self.ystart, self.xend - self.xstart, self.yend - self.ystart])
      self.screenshot_path = os.path.join(application.Server.screenshotPath,
                                          "screenshot-%s.png" % string_tools.get_snowflake_id())
      img.save(self.screenshot_path)
    except Exception:
      pass

  def sys_out(self, event=None):
    if self.center_div:
      self.center_div.destroy()
    self.pos_div.destroy()
    self.root.destroy()


def screenshot():
  st = ScreenshotTools()
  st.start()
  return st.screenshot_path, st.center_offset




from io import BytesIO

from PIL import Image


def get_size(image_path):
  return Image.open(image_path).size

def change_color(image_path, area=None, rgb_color=None):
  """
  修改图片指定区域颜色
  :param image_path: 图片路径
  :param area: 修改区域： (x1, y1, x2, y2)
  :param rgb_color: 入盘颜色 (255, 0, 0)
  :return:
  """
  with Image.open(image_path) as img:
    if area:
      pixels = img.load()
      for x in range(area[0], area[2]):
        for y in range(area[1], area[3]):
          pixels[x, y] = rgb_color
    img_bytes = BytesIO()
    img.save(img_bytes, format='JPEG')
    img_binary = img_bytes.getvalue()
  return img_binary




import base64

from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
from Crypto.Hash import SHA256

from ctools import work_path, cjson

ENCRYPT_CHUNK_SIZE = 245
decrypt_CHUNK_SIZE = 512

def generate_rsa_keypair(bits=2048):
  key = RSA.generate(bits)
  private_key = key.export_key()  # 导出私钥
  public_key = key.publickey().export_key()  # 导出公钥
  with open("private_key.pem", "wb") as f:
    f.write(private_key)
  with open("public_key.pem", "wb") as f:
    f.write(public_key)

def loadLicenseInfo(auth_code):
  with open(work_path.get_app_path() + '/keys/license.key', 'r') as pri:
    decrypt_message = decrypt(auth_code.strip(), pri.read())
    return cjson.loads(decrypt_message)

# 加密函数
def encrypt(msg, public_key):
  parts = b''
  public_key = RSA.import_key(public_key)
  cipher = PKCS1_OAEP.new(public_key)
  for i in range(0, len(msg), ENCRYPT_CHUNK_SIZE):
    parts += cipher.encrypt(msg[i:i + ENCRYPT_CHUNK_SIZE].encode())
  encrypted_base64 = base64.b64encode(parts)
  return encrypted_base64.decode()

# 解密函数
def decrypt(msg, private_key):
  parts = b''
  public_key = RSA.import_key(private_key)
  cipher = PKCS1_OAEP.new(public_key)
  encrypted_bytes = base64.b64decode(msg)
  for i in range(0, len(encrypted_bytes), decrypt_CHUNK_SIZE):
    parts += cipher.decrypt(encrypted_bytes[i:i + decrypt_CHUNK_SIZE])
  return parts.decode()

# 验签
def verify_sign(msg, public_key, sign):
  public_key = RSA.import_key(public_key)
  hash_message = SHA256.new(msg.encode())
  try:
    pkcs1_15.new(public_key).verify(hash_message, base64.b64decode(sign.encode()))
    return True
  except Exception as e:
    print('签名验证失败: {}'.format(e))
    return False

# 签名
def sign_msg(msg, private_key):
  private_key = RSA.import_key(private_key)
  hash_message = SHA256.new(msg.encode())
  signature = pkcs1_15.new(private_key).sign(hash_message)
  return base64.b64encode(signature).decode()


# with open(work_path.get_current_path() + '/private_key.pem', 'r') as key:
#   key = key.read()
#   sign = sign_msg(key, key)
#   with open(work_path.get_current_path() + '/public_key.pem', 'r') as pub:
#     print(verify_sign(key, pub.read(), sign+'123'))



from typing import List


class Alignment:
  def __init__(self, horizontal=None, vertical=None):
    self.horizontal = horizontal
    self.vertical = vertical


class Font:
  def __init__(self, name=None, size: int = None, bold: bool = False, italic: bool = False, underline=None):
    self.name = name
    self.size = size
    self.color = None
    self.bold = bold
    self.italic = italic
    self.underline = underline

  def set_color(self, color):
    if color and color.type == 'rgb':
      self.color = color.rgb[2:]


class Style:

  def __init__(self):
    self.font = None
    self.bg_color = None
    self.alignment: Alignment = Alignment()

  def set_bg_color(self, color):
    if color.type == 'rgb' and color.rgb != "00000000":
      self.bg_color = color.rgb[2:]


class WordCell:
  position: list = []
  value: str = None
  style: Style = None


class WordTable:
  max_rows: int = None
  max_cols: int = None
  cells: List[WordCell] = None
  merged_cells: list = None



# -*- coding: utf-8 -*-
import math

x_pi = 3.14159265358979324 * 3000.0 / 180.0
pi = 3.1415926535897932384626  # π
a = 6378245.0  # 长半轴
ee = 0.00669342162296594323  # 偏心率平方

def gcj02_to_bd09(lng, lat):
    """
    火星坐标系(GCJ-02)转百度坐标系(BD-09)
    谷歌、高德——>百度
    :param lng:火星坐标经度
    :param lat:火星坐标纬度
    :return:
    """
    z = math.sqrt(lng * lng + lat * lat) + 0.00002 * math.sin(lat * x_pi)
    theta = math.atan2(lat, lng) + 0.000003 * math.cos(lng * x_pi)
    bd_lng = z * math.cos(theta) + 0.0065
    bd_lat = z * math.sin(theta) + 0.006
    return [bd_lng, bd_lat]


def bd09_to_gcj02(bd_lon, bd_lat):
    """
    百度坐标系(BD-09)转火星坐标系(GCJ-02)
    百度——>谷歌、高德
    :param bd_lat:百度坐标纬度
    :param bd_lon:百度坐标经度
    :return:转换后的坐标列表形式
    """
    x = bd_lon - 0.0065
    y = bd_lat - 0.006
    z = math.sqrt(x * x + y * y) - 0.00002 * math.sin(y * x_pi)
    theta = math.atan2(y, x) - 0.000003 * math.cos(x * x_pi)
    gg_lng = z * math.cos(theta)
    gg_lat = z * math.sin(theta)
    return [gg_lng, gg_lat]


def wgs84_to_gcj02(lng, lat):
    """
    WGS84转GCJ02(火星坐标系)
    :param lng:WGS84坐标系的经度
    :param lat:WGS84坐标系的纬度
    :return:
    """
    if out_of_china(lng, lat):  # 判断是否在国内
        return [lng, lat]
    dlat = _transformlat(lng - 105.0, lat - 35.0)
    dlng = _transformlng(lng - 105.0, lat - 35.0)
    radlat = lat / 180.0 * pi
    magic = math.sin(radlat)
    magic = 1 - ee * magic * magic
    sqrtmagic = math.sqrt(magic)
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * pi)
    dlng = (dlng * 180.0) / (a / sqrtmagic * math.cos(radlat) * pi)
    mglat = lat + dlat
    mglng = lng + dlng
    return [mglng, mglat]


def gcj02_to_wgs84(lng, lat):
    """
    GCJ02(火星坐标系)转GPS84
    :param lng:火星坐标系的经度
    :param lat:火星坐标系纬度
    :return:
    """
    if out_of_china(lng, lat):
        return [lng, lat]
    dlat = _transformlat(lng - 105.0, lat - 35.0)
    dlng = _transformlng(lng - 105.0, lat - 35.0)
    radlat = lat / 180.0 * pi
    magic = math.sin(radlat)
    magic = 1 - ee * magic * magic
    sqrtmagic = math.sqrt(magic)
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * pi)
    dlng = (dlng * 180.0) / (a / sqrtmagic * math.cos(radlat) * pi)
    mglat = lat + dlat
    mglng = lng + dlng
    return [lng * 2 - mglng, lat * 2 - mglat]


def bd09_to_wgs84(bd_lon, bd_lat):
    lon, lat = bd09_to_gcj02(bd_lon, bd_lat)
    return gcj02_to_wgs84(lon, lat)


def wgs84_to_bd09(lon, lat):
    lon, lat = wgs84_to_gcj02(lon, lat)
    return gcj02_to_bd09(lon, lat)


def _transformlat(lng, lat):
    ret = -100.0 + 2.0 * lng + 3.0 * lat + 0.2 * lat * lat + \
          0.1 * lng * lat + 0.2 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 *
            math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lat * pi) + 40.0 *
            math.sin(lat / 3.0 * pi)) * 2.0 / 3.0
    ret += (160.0 * math.sin(lat / 12.0 * pi) + 320 *
            math.sin(lat * pi / 30.0)) * 2.0 / 3.0
    return ret


def _transformlng(lng, lat):
    ret = 300.0 + lng + 2.0 * lat + 0.1 * lng * lng + \
          0.1 * lng * lat + 0.1 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 *
            math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lng * pi) + 40.0 *
            math.sin(lng / 3.0 * pi)) * 2.0 / 3.0
    ret += (150.0 * math.sin(lng / 12.0 * pi) + 300.0 *
            math.sin(lng / 30.0 * pi)) * 2.0 / 3.0
    return ret


def out_of_china(lng, lat):
    """
    判断是否在国内，不在国内不做偏移
    :param lng:
    :param lat:
    :return:
    """
    return not (lng > 73.66 and lng < 135.05 and lat > 3.86 and lat < 53.55)




import tkinter as tk

"""
规划区域工具类, 按回车获取区域坐标
image_path=screenshot_tools.screenshot()
"""


class PlanAreaTools:
  def __init__(self):
    self.root = tk.Tk()
    self.root.overrideredirect(True)
    self.root.attributes("-alpha", 0.1)
    self.root.attributes('-topmost', 'true')
    self.root.geometry("{0}x{1}+0+0".format(self.root.winfo_screenwidth(), self.root.winfo_screenheight()))
    self.root.configure(bg="black")

    self.canvas = tk.Canvas(self.root)

    self.x, self.y = 0, 0
    self.xstart, self.ystart = 0, 0
    self.xend, self.yend = 0, 0
    self.xcenter, self.ycenter = 0, 0

    self.select_area = self.canvas.create_rectangle(0, 0, 0, 0, outline='red', width=0, dash=(4, 4))
    self.center_offset = (0, 0)

    self.pos_div = tk.Toplevel(self.root)
    self.pos_div.overrideredirect(True)
    self.pos_div.attributes('-topmost', 'true')
    self.pos_div.configure(bg="grey")
    self.pos_div.geometry("270x25")

    self.pos_div_text = "POS: (%s, %s), Enter键-获取区域, Esc键-退出"
    self.pos_label = tk.Label(self.pos_div, text=self.pos_div_text % (0, 0), bg="black", fg="white")
    self.pos_label.pack(fill="both", expand=True)

    self.root.bind('<Escape>', self.sys_out)  # 键盘Esc键->退出
    self.root.bind('<Motion>', self.print_pos)
    self.root.bind("<Button-1>", self.mouse_left_down)  # 鼠标左键点击->显示子窗口
    self.root.bind("<B1-Motion>", self.mouse_left_move)  # 鼠标左键移动->改变子窗口大小
    self.root.bind("<ButtonRelease-1>", self.mouse_left_up)  # 鼠标左键释放->记录最后光标的位置
    self.root.bind("<Return>", self.save_area)  # 回车键->获取区域信息
    self.root.focus()

    self.area = [0, 0, 0, 0]

  def start(self):
    self.root.mainloop()

  def print_pos(self, event):
    self.x, self.y = event.widget.winfo_pointerxy()
    self.pos_label.config(text=self.pos_div_text % (self.x, self.y))
    self.pos_div.geometry(f"+{self.x + 10}+{self.y + 10}")

  def mouse_left_down(self, event):
    self.x, self.y = event.x, event.y
    self.xstart, self.ystart = event.x, event.y
    self.canvas.configure(height=1)
    self.canvas.configure(width=1)
    self.canvas.config(highlightthickness=0)
    self.canvas.place(x=event.x, y=event.y)

  def mouse_left_move(self, event):
    self.x, self.y = event.x, event.y
    self.xcenter = self.xstart + int((self.x-self.xstart)/2)-2
    self.ycenter = self.ystart + int((self.y-self.ystart)/2)-2

    self.pos_label.config(text=self.pos_div_text % (self.x, self.y))
    self.pos_div.geometry(f"+{self.x + 10}+{self.y + 10}")

    self.canvas.configure(height=event.y - self.ystart)
    self.canvas.configure(width=event.x - self.xstart)
    self.canvas.coords(self.select_area, 0, 0, event.x - self.xstart, event.y - self.ystart)

  def mouse_left_up(self, event):
    self.xend, self.yend = event.x, event.y

  def save_area(self, event):
    try:
      self.area = [self.xstart, self.ystart, self.xend, self.yend]
      self.canvas.delete(self.select_area)
      self.canvas.place_forget()
      self.sys_out()

    except Exception:
      pass

  def sys_out(self, event=None):
    self.pos_div.destroy()
    self.root.destroy()


def get_area():
  pat = PlanAreaTools()
  pat.start()
  return tuple(pat.area)



from cryptography.fernet import Fernet

def generate_key():
  """
  生成 AES key
  Returns 32 bytes key
  -------
  """
  key = Fernet.generate_key()
  return key.decode()

def aes_encrypt(sec_key, plaintext):
  """
  aes加密
  :param sec_key: 加密 key
  :param plaintext: 明文信息
  :return: 加密后的信息
  """
  cipher = Fernet(sec_key)
  ciphertext = cipher.encrypt(plaintext.encode())
  return ciphertext.decode()


def aes_decrypt(sec_key, ciphertext):
  """
  aes解密
  :param sec_key: 加密 key
  :param ciphertext: 密文
  :return: 解密后的明文信息
  """
  cipher = Fernet(sec_key)
  decrypted_data = cipher.decrypt(ciphertext)
  return decrypted_data.decode()





from ctools.snow_id import SnowId

idWorker = SnowId(1, 2, 0)

def get_random_str(size: int = 10) -> str:
  import random
  return "".join(random.sample('abcdefghjklmnpqrstuvwxyz123456789', size))

def get_uuid() -> str:
  import uuid
  return str(uuid.uuid1()).replace("-", "")

def get_snowflake_id():
  return idWorker.get_id()

def decode_bytes(bytes_str):
  import chardet
  res_str = ""
  if bytes_str:
    detect = chardet.detect(bytes_str)
    if detect:
      confidence = 0
      chardet_error = False
      try:
        confidence = float(detect.get('confidence'))
        res_str = bytes_str.decode(encoding=detect.get('encoding'))
      except Exception:
        chardet_error = True

      try:
        if confidence <= 0.95 or chardet_error:
          encoding = "utf-8" if detect.get('encoding') == "utf-8" else "gbk"
          res_str = bytes_str.decode(encoding=encoding)
      except Exception:
        res_str = str(bytes_str)
  return res_str


def check_sum(content: str):
  import hashlib
  try:
    algorithm = hashlib.sha256()
    algorithm.update(content.encode())
    return algorithm.hexdigest()
  except Exception:
    return None


def is_list(v: str):
  try:
    list(v)
    if v[0] == "[" and v[-1] == "]":
      return True
    else:
      return False
  except Exception:
    return False

def is_digit(v: str):
  try:
    float(v)
    return True
  except Exception:
    return False

def is_bool(v: str):
  if v in ["False", "True"]:
    return True
  else:
    return False

def dict_to_params(obj: dict):
  params = ""
  for k, v in obj.items():
    if k == 'varname':
      continue
    v = str(v)
    if not is_list(v) and not is_digit(v) and not is_bool(v):
      if k == "path" and v[:4] != "http":
        v = "r'%s'" % v
      else:
        v = "'%s'" % v
    params += "%s=%s, " % (k, v)
  params = params[:params.rfind(',')]
  return params



import requests


def get(url, params=None, headers=None):
  result = ""
  try:
    response = requests.get(url, params=params, headers=headers, timeout=60, verify=False)
    response.raise_for_status()
    if response.status_code == 200:
      result = response.content
  except Exception as e:
    print("GET请求异常:", e)
  if isinstance(result, bytes): return result.decode('utf-8')
  return result


def post(url, data=None, json=None, headers=None, files=None):
  result = ""
  response = requests.post(url, data=data, json=json, files=files, headers=headers, timeout=60, verify=False)
  response.raise_for_status()
  if response.status_code == 200:
    result = response.content
  if isinstance(result, bytes): return result.decode('utf-8')
  return result



import os
import shutil

from sqlalchemy.sql import text

from ctools import database


class Patch:

  def __init__(self, oldVersion, newVersion, pythonPath, playwrightPath, driverPath) -> None:
    super().__init__()
    self.oldV = oldVersion[len('V'):].replace('.', '').replace('-snapshot', '')
    self.currentV = newVersion[len('V'):].replace('.', '').replace('-snapshot', '')
    self.snapshot = '-snapshot' in newVersion or '-snapshot' in oldVersion
    self.pythonPath = pythonPath
    self.playwrightPath = playwrightPath
    self.driverPath = driverPath
    self.deleteSDK, self.deleteHPL, self.deleteDriver = False, False, False

  def apply_patch(self):
    patch_methods = [method for method in dir(self) if callable(getattr(self, method)) and method.startswith('patch_V')]
    patch_methods.sort(key=lambda x: int(x[len('patch_V'):]))
    max_method_name = patch_methods[-1]
    exec_max_method = False
    for method_name in patch_methods:
      slVersion = method_name[len('patch_V'):]
      if int(self.currentV) > int(slVersion) >= int(self.oldV):
        if max_method_name == method_name: exec_max_method = True
        method = getattr(self, method_name)
        print('start exec patch {}'.format(method_name))
        method()
        print('patch {} update success'.format(method_name))
    if self.snapshot and not exec_max_method:
      print('start exec snapshot patch {}'.format(max_method_name))
      method = getattr(self, max_method_name)
      method()
      print('snapshot patch {} update success'.format(max_method_name))

  def patch_V220(self):
    pass

  def run_sqls(self, sqls):
    with database.get_session() as s:
      for sql in sqls.split(";"):
        try:
          s.execute(text(sql.strip()))
          s.commit()
        except Exception as e:
          print('结构升级错误, 请检查!!! {}'.format(e.__cause__))

  def remove_sdk(self):
    if self.deleteSDK: return
    try:
      self.deleteSDK = True
      if os.path.exists(self.pythonPath): shutil.rmtree(self.pythonPath)
    except Exception as e:
      print('删除SDK错误: {}'.format(e))

  def remove_hpl(self):
    if self.deleteHPL: return
    try:
      self.deleteHPL = True
      if os.path.exists(self.playwrightPath): shutil.rmtree(self.playwrightPath)
    except Exception as e:
      print('删除HPL错误: {}'.format(e))

  def remove_driver(self):
    if self.deleteDriver: return
    try:
      self.deleteDriver = True
      if os.path.exists(self.driverPath): shutil.rmtree(self.driverPath)
    except Exception as e:
      print('删除Driver错误: {}'.format(e))



# from docxtpl import DocxTemplate
#
# # tpl = DocxTemplate('/Users/haoyang/Desktop/xxx.docx')
# tpl = DocxTemplate('/Users/haoyang/Desktop/123.doc')
#
# # 设置好各标签需要填写的内容
# context = {'xxxx': '计算机科学与技术', 'cccc': '2022050513'}
# # 将标签内容填入模板中
# tpl.render(context)
# # 保存
# tpl.save('/Users/haoyang/Desktop/new_test2.docx')

from docx import Document
def merge_word_files(input_files, output_file):
    merged_doc = Document()
    for file in input_files:
        doc = Document(file)
        for element in doc.element.body:
            merged_doc.element.body.append(element)
    merged_doc.save(output_file)
# 示例用法
input_files = ["/Users/haoyang/Desktop/xxx.docx", "/Users/haoyang/Desktop/xxx 2.docx"]
output_file = "/Users/haoyang/Desktop/merged_file.docx"
merge_word_files(input_files, output_file)



from ctools import cjson

cjson.str_value_keys = [
  "obj_id",
]


class _ResEnum(object):
  def __init__(self, code: int, message: str):
    self.code = code
    self.message = message

  def __eq__(self, o: object) -> bool:
    return self.code == o

class R(object):
  class Code:

    @staticmethod
    def cus_code(code, msg):
      return _ResEnum(code, msg)

    SUCCESS = _ResEnum(200, "成功")
    FAIL    = _ResEnum(400, "失败")
    ERROR   = _ResEnum(500, "异常")

  def __init__(self, code: int, message: str, data=""):
    self.code = code
    self.message = message
    self.data = data

  def _to_json_str(self):
    return cjson.unify_to_str(cjson.dumps(self))

  @staticmethod
  def parser(r_json: str):
    return R(**cjson.loads(r_json))

  @staticmethod
  def ok(data=None, resp=Code.SUCCESS, msg=None, to_json_str=True):
    if not to_json_str:
      return R(resp.code, msg if msg is not None else resp.message, data)
    return R(resp.code, msg if msg is not None else resp.message, data)._to_json_str()

  @staticmethod
  def fail(msg=None, resp=Code.FAIL, data=None, to_json_str=True):
    if not to_json_str:
      return R(resp.code, msg if msg is not None else resp.message, data)
    return R(resp.code, msg if msg is not None else resp.message, data)._to_json_str()

  @staticmethod
  def error(msg=None, resp=Code.ERROR, data=None, to_json_str=True):
    if not to_json_str:
      return R(resp.code, msg if msg is not None else resp.message, data)
    return R(resp.code, msg if msg is not None else resp.message, data)._to_json_str()



import importlib
import os
import time


def code_to_pyc(code: str, out_file_path: str):
  file_name = os.path.split(out_file_path)[-1]
  compiled_code = compile(code, file_name, 'exec')
  bytecode = importlib._bootstrap_external._code_to_timestamp_pyc(compiled_code, time.time(), len(code))
  with open(out_file_path, 'wb') as f:
    f.write(bytecode)


def file_to_pyc(file_path: str, out_file_path: str):
  with open(file_path, 'r') as f:
    code = f.read()
  if code:
    code_to_pyc(code, out_file_path)



import base64

from gmssl import sm2
from gmssl.sm4 import CryptSM4, SM4_ENCRYPT, SM4_DECRYPT

sm2_crypt: sm2.CryptSM2 = None

def init(private_key: str, public_key: str):
  global sm2_crypt
  if sm2_crypt is not None:
    print('sm2 is already init!!!')
    return
  sm2_crypt = sm2.CryptSM2(private_key=private_key, public_key=public_key, asn1=True, mode=1)

def sign_with_sm2(sign_data: str) -> str:
  if sm2_crypt is None: raise Exception('sm2 is not init!!!')
  return sm2_crypt.sign_with_sm3(sign_data.encode('UTF-8'))


def verify_with_sm2(sign_val: str, sign_data: str) -> bool:
  if sm2_crypt is None: raise Exception('sm2 is not init!!!')
  try:
    return sm2_crypt.verify_with_sm3(sign_val, sign_data.encode('UTF-8'))
  except Exception as e:
    print('签名验证失败: {}'.format(e))
    return False

def encrypt_with_sm2(encrypt_data: str) -> str:
  if sm2_crypt is None: raise Exception('sm2 is not init!!!')
  return base64.b64encode(sm2_crypt.encrypt(encrypt_data.encode('UTF-8'))).decode('UTF-8')


def decrypt_with_sm2(encrypt_data: str) -> str:
  if sm2_crypt is None: raise Exception('sm2 is not init!!!')
  return sm2_crypt.decrypt(base64.b64decode(encrypt_data.encode('UTF-8'))).decode('UTF-8')


def encrypt_with_sm4(key: bytes, encrypt_text: str):
  crypt_sm4 = CryptSM4()
  crypt_sm4.set_key(key, SM4_ENCRYPT)
  encrypt_value = base64.b64encode(crypt_sm4.crypt_ecb(encrypt_text.encode()))
  return encrypt_value.decode()


def decrypt_with_sm4(key: bytes, decrypt_text: str):
  crypt_sm4 = CryptSM4()
  crypt_sm4.set_key(key, SM4_DECRYPT)
  decrypt_value = crypt_sm4.crypt_ecb(base64.b64decode(decrypt_text))
  return decrypt_value.decode()




def isNull(param) -> bool:
    if type(param) == str:
        return param == ''
    elif type(param) == list:
        return len(param) == 0
    elif type(param) == dict:
        return len(param) == 0
    elif type(param) == int:
        return param == 0
    elif type(param) == float:
        return param == 0.0
    elif type(param) == bool:
        return param == False
    else:
        return param is None

def isNotNull(param)  -> bool:
    return not isNull(param)




import os
import threading
import time
from concurrent.futures import ThreadPoolExecutor

from ctools import call

thread_local = threading.local()

_threadPool: ThreadPoolExecutor = None

@call.init
def init():
  global _threadPool
  max_work_num = min(32, (os.cpu_count() or 1) + 4) # 最多 32 个
  _threadPool = ThreadPoolExecutor(max_workers= max_work_num, thread_name_prefix='ck-')

def cb(f, callback):
  exc = f.exception()
  if exc:
    if callback: callback(exc)
    raise exc
  else:
    if callback: callback(f.result())

def submit(func, *args, callback=None, **kwargs):
  if _threadPool is None: raise Exception('thread pool is not init')
  future = _threadPool.submit(func, *args, **kwargs)
  future.add_done_callback(lambda f: cb(f, callback))
  time.sleep(0.01)
  return future

def shutdown(wait=True):
  if _threadPool is None: raise Exception('thread pool is not init')
  _threadPool.shutdown(wait=wait)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2025/1/21 16:01'

import time
import jwt
from bottle import request

from ctools.dict_wrapper import DictWrapper

token_header = 'Authorization'

def gen_token(payload: {}, secret_key, expired: int=3600) -> str:
  payload.update({'exp': time.time() + expired})
  return jwt.encode(payload, secret_key, algorithm='HS256')

def get_payload(token, secret_key):
  try:
    payload = jwt.decode(token, secret_key, algorithms=['HS256'])
    return DictWrapper(payload)
  except Exception as e:
    return None

def get_token(key):
  return get_payload(request.get_header(token_header), key)

def is_valid(key):
  return get_payload(request.get_header(token_header), key) is not None

# if __name__ == '__main__':
#   token = gen_token({"xx": 123}, '123')
#   xx = get_payload(token, '123')
#   print(xx.xx)



import os
import time

import pyautogui
import uiautomation as auto
from pynput import keyboard

from ctools import thread_pool, win_canvas, application, string_tools

current_control = None
ctrl_pressed = False
keyboard_listener = None
control_json = {}
screenshot_path = ""

class ControlInfo:
  def __init__(self, ControlType, ClassName, AutomationId, Name, Depth):
    self.ControlType = ControlType
    self.ClassName = ClassName
    self.AutomationId = AutomationId
    self.Name = Name
    #self.Depth = Depth


def get_control_from_cursor():
  global control_json, screenshot_path
  with auto.UIAutomationInitializerInThread():
    win_canvas.data_queue.put(win_canvas.CanvasInfo(msg_type=win_canvas.MSGType.BOTTOM))
    time.sleep(0.2)
    try:
      control = auto.ControlFromCursor()
    except Exception as e:
      print("No control found {}".format(e))
      return
    if control:
      #当前控件信息
      global current_control
      current_control = control
      #绘制矩形
      canvas_info = win_canvas.CanvasInfo(control.Name, [
        (control.BoundingRectangle.left, control.BoundingRectangle.top),
        (control.BoundingRectangle.right, control.BoundingRectangle.top),
        (control.BoundingRectangle.right, control.BoundingRectangle.bottom),
        (control.BoundingRectangle.left, control.BoundingRectangle.bottom)
      ])
      win_canvas.data_queue.put(canvas_info)
      control_json = {}
      c_info = ControlInfo(ControlType=current_control.ControlType, ClassName=current_control.ClassName, AutomationId=current_control.AutomationId, Name=current_control.Name, Depth=0)
      _depth = 0
      while current_control:
        current_control = current_control.GetParentControl()
        if current_control: _depth += 1
      c_info.Depth = _depth
      control_json.update(c_info.__dict__)

      img = pyautogui.screenshot(region=[control.BoundingRectangle.left, control.BoundingRectangle.top,
                                         control.BoundingRectangle.width(), control.BoundingRectangle.height()])
      screenshot_path = os.path.join(application.Server.screenshotPath, "screenshot-%s.png" % string_tools.get_snowflake_id())
      img.save(screenshot_path)
      # xx = auto.Control(**control_json)
      # print(control_json)
      # time.sleep(2)
      # xx.Click()

def on_press(key):
  global ctrl_pressed, keyboard_listener
  if key == keyboard.Key.ctrl_l and not ctrl_pressed:
    ctrl_pressed = True
    thread_pool.submit(get_control_from_cursor)
  elif key == keyboard.Key.esc:
    win_canvas.stop()
    keyboard_listener.stop()
  elif hasattr(key, 'vk') and key.vk == 192 and ctrl_pressed:
    win_canvas.stop()
    keyboard_listener.stop()
    # pg.alert('采集成功!')

def on_release(key):
  global ctrl_pressed
  if key == keyboard.Key.ctrl_l:
    ctrl_pressed = False
    win_canvas.data_queue.put(win_canvas.CanvasInfo(msg_type=win_canvas.MSGType.BOTTOM))


def start():
  global keyboard_listener
  win_canvas.start()
  keyboard_listener = keyboard.Listener(on_press=on_press, on_release=on_release)
  keyboard_listener.start()
  keyboard_listener.join()


def get_control_json():
  global current_control, ctrl_pressed, keyboard_listener, control_json
  current_control = None
  ctrl_pressed = False
  keyboard_listener = None
  control_json = {}
  start()
  if len(control_json) == 0:
    time.sleep(0.5)
  return control_json, screenshot_path

if __name__ == '__main__':
  a = get_control_json()
  print(a)



import os
import threading
from enum import Enum

from prometheus_client import Counter, Gauge, Summary, Histogram, start_http_server
from ctools import call, cjson, sys_log, work_path

log = sys_log.flog

_metrics_port = 8889
_persistent_json = {}
_metrics_initial = {}
_lock = threading.Lock()
_metrics_persistent_path = os.path.join(work_path.get_user_work_path('metrics', True), 'persistent.json')

# 认证中间件
# @app.before_request
# def check_authentication():
#   auth = request.authorization
#   if not auth or auth.username != USERNAME or auth.password != PASSWORD:
#     return Response(
#       "Unauthorized", 401, {"WWW-Authenticate": 'Basic realm="Login Required"'}
#     )

class MetricType(Enum):
  COUNTER = 'counter'
  GAUGE = 'gauge'
  SUMMARY = 'summary'
  HISTOGRAM = 'histogram'

class Metric:
  def __init__(self, metric_type: MetricType, metric_key: str, metric_labels: [], persistent: bool = False, buckets: [] = None, reset: bool = False, desc: str = ""):
    self.metric_type = metric_type
    self.metric_key = metric_key
    self.metric_labels = metric_labels
    self.buckets = buckets
    self.metric = None
    self.persistent = persistent
    self.reset = reset
    if metric_type == MetricType.COUNTER:
      self.metric = Counter(metric_key, desc, metric_labels)
    elif metric_type == MetricType.GAUGE:
      self.metric = Gauge(metric_key, desc, metric_labels)
    elif metric_type == MetricType.SUMMARY:
      self.metric = Summary(metric_key, desc, metric_labels)
    elif metric_type == MetricType.HISTOGRAM:
      if buckets is None: raise Exception('histogram buckets can not empty')
      self.metric = Histogram(metric_key, desc, metric_labels, buckets=self.buckets)
    else:
      raise Exception('metric type not found')
    _metrics_initial[metric_key] = self

@call.once
def init(reset_persistent: bool = False):
  if os.path.exists(_metrics_persistent_path) and not reset_persistent:
    with open(_metrics_persistent_path, 'r') as persistent_file:
      global _persistent_json
      try:
        content = persistent_file.readline()
        log.info("persistent初始化: %s" % content)
        _persistent_json = cjson.loads(content)
      except Exception:
        log.error('persistent.json is not valid json!!!!!')
        _persistent_json = {}
  _init_all_metrics()
  _persistent_json = _persistent_json or {}
  for key, item in _persistent_json.items():
    metrics_key = key.split("-")[0]
    if '_labels' in key or metrics_key not in _metrics_initial: continue
    opt(metrics_key, _persistent_json[key + '_labels'], _persistent_json[key])
  persistent_metrics()
  start_http_server(port=_metrics_port)

@call.schd(5, start_by_call=True)
def persistent_metrics():
  if _persistent_json and not _lock.locked():
    log.info('begin persistent metrics to file...')
    with open(_metrics_persistent_path, 'w') as persistent_file:
      persistent_file.write(cjson.dumps(_persistent_json))
      persistent_file.flush()

def opt(metric_key: str, label_values: [], metric_value: int):
  _lock.acquire(timeout=5)
  try:
    persistent_key = "%s-%s" % (metric_key, "_".join(map(str, label_values)))
    metric_entity: Metric = _metrics_initial[metric_key]
    if metric_entity.persistent:
      if not metric_entity.reset and persistent_key in _persistent_json:
        _persistent_json[persistent_key] += metric_value
      else:
        _persistent_json[persistent_key] = metric_value
      _persistent_json[persistent_key + '_labels'] = label_values

      if _persistent_json[persistent_key] < 0:
        _persistent_json[persistent_key] = 0
        metric_value = 0

    if metric_entity.metric_type == MetricType.COUNTER or metric_entity.metric_type == MetricType.GAUGE:
      if label_values is None or len(label_values) == 0:
        if metric_entity.metric_type == MetricType.COUNTER and metric_entity.reset:
          metric_entity.metric.labels('').reset()
        if metric_entity.metric_type == MetricType.GAUGE and metric_entity.reset:
          metric_entity.metric.labels('').set(0)
        metric_entity.metric.labels('').inc(metric_value)
      else:
        if metric_entity.reset:
          if metric_entity.metric_type == MetricType.COUNTER and metric_entity.reset:
            metric_entity.metric.labels(*label_values).reset()
          if metric_entity.metric_type == MetricType.GAUGE and metric_entity.reset:
            metric_entity.metric.labels(*label_values).set(0)
        metric_entity.metric.labels(*label_values).inc(metric_value)
    else:
      if label_values is None or len(label_values) == 0:
        metric_entity.metric.observe(metric_value)
      else:
        metric_entity.metric.labels(*label_values).observe(metric_value)
  except Exception as e:
    log.error("添加指标信息异常: %s" % e)
  _lock.release()

def _init_all_metrics():
  Metric(MetricType.GAUGE, 'gomyck', ['g_label1', 'g_label2'], persistent=True, reset=True)

# if __name__ == '__main__':
#   init()
#   import random
#   while True:
#     opt('data_reported_time', ['123', '123'], random.randint(1, 10))
#     opt('data_received_time', ['123', '123'], random.randint(1, 10))
#     opt('data_insert_time',   ['123', '123'], random.randint(1, 10))
#     time.sleep(1)



import sys
from socketserver import ThreadingMixIn
from wsgiref.simple_server import WSGIServer, WSGIRequestHandler, make_server

from bottle import ServerAdapter, Bottle, template, static_file, abort, redirect, response

from ctools import sys_info

"""
module_names = list(globals().keys())
def get_modules():
  mods = []
  for modname in module_names:
    if modname == 'base' or modname == 'online' or modname.startswith('__') or modname == 'importlib': continue
    module = globals()[modname]
    mods.append(module)
  return mods

def get_ws_modules():
  from . import websocket
  return [websocket]
"""

"""
from ctools import bottle_web_base, token, bottle_webserver
from ctools.api_result import R

secret_key = "xxx"
app = bottle_web_base.init_app('子模块写 context_path, 主模块就不用写任何东西')

# 通用的鉴权方法
@bottle_web_base.before_intercept(0)
def token_check():
  return bottle_web_base.common_auth_verify(secret_key)

@app.post('/login')
def login(params):
  return R.ok(token.gen_token({'username': 'xxx'}, secret_key, 3600))

@app.get('/demo')
@bottle_web_base.rule('DOC:DOWNLOAD')
def demo(params):
  print(params)

main_app = bottle_webserver.init_bottle() # 这里可以传 APP 当做主模块, 但是 context_path 就不好使了, 上下文必须是 /
main_app.mount(app.context_path, app)
main_app.set_index(r'轨迹点位压缩.html')
main_app.run()
"""

_default_port = 8888

class CBottle:

  def __init__(self, bottle: Bottle, port=_default_port, quiet=False):
    self.port = port
    self.quiet = quiet
    self.bottle = bottle
    self.index_root = './'
    self.index_filename = 'index.html'
    self.is_tpl = False
    self.tmp_args = {}
    self.redirect_url = None
    self.static_root = './static'
    self.download_root = './download'

    @self.bottle.route(['/', '/index'])
    def index():
      try:
        if self.redirect_url: return redirect(self.redirect_url)
        if self.is_tpl: return template(f"{self.index_root}/{self.index_filename}", self.tmp_args)
        return static_file(filename=self.index_filename, root=self.index_root)
      except FileNotFoundError:
        abort(404, "File not found...")

    @self.bottle.route('/static/<filepath:path>')
    def static(filepath):
      try:
        return static_file(filepath, root=self.static_root)
      except FileNotFoundError:
        abort(404, "File not found...")

    @self.bottle.route('/download/<filepath:path>')
    def download(filepath):
      return static_file(filepath, root=self.download_root, download=True)

    @self.bottle.route('/favicon.ico')
    def favicon():
      response.content_type = 'image/svg+xml'
      svg_icon = '''<?xml version="1.0" encoding="UTF-8"?>
      <svg width="32" height="32" viewBox="0 0 32 32" xmlns="http://www.w3.org/2000/svg">
          <circle cx="16" cy="16" r="14" fill="#007bff"/>
          <path d="M16 8a8 8 0 0 0-8 8h2a6 6 0 0 1 12 0h2a8 8 0 0 0-8-8z" fill="white"/>
          <circle cx="16" cy="20" r="2" fill="white"/>
      </svg>
      '''
      return svg_icon

  def run(self):
    http_server = WSGIRefServer(port=self.port)
    print('Click the link below to open the service homepage %s' % '\n \t\t http://localhost:%s \n \t\t http://%s:%s' %  (self.port, sys_info.get_local_ipv4(), self.port), file=sys.stderr)
    self.bottle.run(server=http_server, quiet=self.quiet)

  def set_index(self, filename='index.html', root='./', is_tpl=False, redirect_url=None, **kwargs):
    self.index_root = root
    self.index_filename = filename
    self.is_tpl = is_tpl
    self.redirect_url = redirect_url
    self.tmp_args = kwargs

  def set_static(self, root='./static'):
    self.static_root = root

  def set_download(self, root='./download'):
    self.download_root = root

  def mount(self, context_path, app, **kwargs):
    self.bottle.mount(context_path, app, **kwargs)

def init_bottle(app:Bottle=None, port=_default_port, quiet=False) -> CBottle:
  bottle = app or Bottle()
  return CBottle(bottle, port, quiet)

class ThreadedWSGIServer(ThreadingMixIn, WSGIServer):
  daemon_threads = True

class CustomWSGIHandler(WSGIRequestHandler):
  def log_request(*args, **kw): pass

class WSGIRefServer(ServerAdapter):

  def __init__(self, host='0.0.0.0', port=_default_port):
    super().__init__(host, port)
    self.server = None

  def run(self, handler):
    req_handler = WSGIRequestHandler
    if self.quiet: req_handler = CustomWSGIHandler
    self.server = make_server(self.host, self.port, handler, server_class=ThreadedWSGIServer, handler_class=req_handler)
    self.server.serve_forever()

  def stop(self):
    self.server.shutdown()



import time
import traceback
from functools import wraps

# annotation
def exception_handler(fail_return, retry_num=0, delay=3, catch_e=Exception, print_exc=False):
  def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      try:
        return func(*args, **kwargs)
      except catch_e as e:
        print(f"{func.__name__} runtime exception: {str(e)}")
        if print_exc: traceback.print_exc()
        nonlocal retry_num
        renum = retry_num
        if renum == 0:
          return fail_return
        else:
          while renum > 0:
            time.sleep(delay)
            renum -= 1
            try:
              return func(*args, **kwargs)
            except catch_e:
              pass
          return fail_return

    return wrapper

  return decorator



import hashlib
import io
import json
import os
import shutil
import zipfile
from base64 import b64decode
from base64 import b64encode

from Crypto.Hash import MD5
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15

from ctools import sys_log, work_path, application

log = sys_log.flog


def data_layout(manifest: dict):
  """
  数据重新排版
  :param manifest:
  :return:
  """
  data = ""
  for key in sorted(manifest):
    if key != "signature":
      data += "%s=%s," % (key, manifest[key])
  return data[:-1]


def get_md5_sum(data):
  return hashlib.md5(data + "gomyck-daning".encode()).hexdigest()


def signature(manifest: dict):
  """
  定义签名函数，能够使用指定的私钥对数据文件进行签名，并将签名结果输出到文件返回
  :param manifest: 元数据字典
  :return:
  """
  with open(work_path.get_app_path() + '/keys/resource_private_key.pem', 'r') as pri:
    private_key = RSA.import_key(pri.read())
    digest = MD5.new(data_layout(manifest).encode('utf-8'))
    return b64encode(pkcs1_15.new(private_key).sign(digest)).decode()


def sign_verify(manifest: dict, sign_val: str):
  """
  定义签名验证函数，能够使用指定的公钥对任务2中的签名文件进行验证，返回验证结果
  :param manifest: 元数据字典
  :param sign_val: 对比的签名值
  :return:
  """
  with open(work_path.get_app_path() + '/keys/resource_public_key.pem', 'r') as pub:
    public_key = RSA.import_key(pub.read())
    digest = MD5.new(data_layout(manifest).encode('utf-8'))
    try:
      pkcs1_15.new(public_key).verify(digest, b64decode(sign_val))
      return True
    except:
      return False


def unpack_resource(resource_name, file):
  """
  解压资源包并验证包的完整性和签名
  :param resource_name:
  :param file:
  :return:
  """
  message = None
  file_md5_sum = os.path.splitext(resource_name)[0].split('-')[-1]
  if file_md5_sum != get_md5_sum(file.read()):
    message = "资源包已损坏, 请重新打包"
  else:
    manifest = {}
    with zipfile.ZipFile(file, 'r') as zf:
      for file in zf.filelist:
        if file.filename == "manifest.json":
          manifest = json.loads(zf.read(file.filename))
          break
      # 校验签名
      if sign_verify(manifest, manifest.get('signature')):
        resource_path = os.path.join(application.Upload.source_pkg_path, manifest['name'])
        shutil.rmtree(resource_path, ignore_errors=True)
        zf.extractall(resource_path)
      else:
        message = "资源包签名校验未通过"
  return message


def pack_resource(resource_path):
  """
  压缩打包资源包为zip文件
  :param resource_path:
  :return:
  """
  file_name = None
  buffer = io.BytesIO()
  manifest_path = os.path.join(resource_path, 'manifest.json')
  if os.path.exists(manifest_path):
    resource_name = os.path.split(resource_path)[-1]
    with open(manifest_path, mode="r", encoding='utf-8') as f:
      manifest = json.loads(f.read())
      manifest["signature"] = signature(manifest)
    with open(manifest_path, mode="w", encoding='utf-8') as f:
      f.write(json.dumps(manifest, indent=2, ensure_ascii=False))

    with zipfile.ZipFile(buffer, 'w', zipfile.ZIP_DEFLATED, allowZip64=False) as zf:
      for dir_path, dir_names, file_names in os.walk(resource_path):
        for filename in file_names:
          filepath = os.path.join(dir_path, filename)
          arcname = os.path.join(dir_path.split(resource_name)[-1], filename)
          zf.write(filepath, arcname=arcname)
    buffer.seek(0)
    md5_sum = get_md5_sum(buffer.getvalue())
    file_name = "%s-%s.zip" % (resource_name, md5_sum)
  else:
    log.info("资源包%s不存在manifest.json元数据文件， 无法打包" % resource_path)
  return file_name, buffer



import logging
import os
import sys
import time
import traceback

from ctools import call, work_path

clog: logging.Logger = None
flog: logging.Logger = None

neglect_keywords = [
  "OPTIONS",
]


# 文件日志
@call.once
def _file_log(sys_log_path: str = './', log_level: int = logging.INFO, mixin: bool = False) -> logging:
  try:
    os.mkdir(sys_log_path)
  except Exception:
    pass
  log_file = sys_log_path + os.path.sep + "log-" + time.strftime("%Y-%m-%d-%H", time.localtime(time.time())) + ".log"
  if mixin:
    handlers = [logging.FileHandler(filename=log_file, encoding='utf-8'), logging.StreamHandler()]
  else:
    handlers = [logging.FileHandler(filename=log_file, encoding='utf-8')]
  logging.basicConfig(level=log_level,
                      format='%(asctime)s-%(levelname)s-%(thread)d-%(module)s(%(funcName)s:%(lineno)d) %(message)s',
                      datefmt='%Y%m%d%H%M%S',
                      handlers=handlers)
  logger = logging.getLogger('ck-flog')
  return logger


# 控制台日志
@call.once
def _console_log(log_level: int = logging.INFO) -> logging:
  handler = logging.StreamHandler()
  logging.basicConfig(level=log_level,
                      format='%(asctime)s-%(levelname)s-%(thread)d-%(module)s(%(funcName)s:%(lineno)d) %(message)s',
                      datefmt='%Y%m%d%H%M%S',
                      handlers=[handler])
  logger = logging.getLogger('ck-clog')
  return logger


class GlobalLogger(object):
  def __init__(self, logger):
    sys.stdout = self
    sys.stderr = self
    sys.excepthook = self.handle_exception
    self.log = logger

  def write(self, message):
    if message == '\n' or message == '\r\n': return
    global neglect_keywords
    for neglect_keyword in neglect_keywords:
      if neglect_keyword in message: return
    try:
      stack = traceback.extract_stack(limit=3)
      caller = stack[-2]
      location = f"{os.path.splitext(os.path.basename(caller.filename))[0]}({caller.name}:{caller.lineno})"
      self.log.info(f"{location} {message.strip()}")
    except Exception:
      self.log.info(message.strip())

  def handle_exception(self, exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
      sys.__excepthook__(exc_type, exc_value, exc_traceback)
      return
    formatted_exception = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    self.log.error(f"An error occurred:\n{formatted_exception.strip()}")

  def flush(self):
    pass


@call.init
def _init_log() -> None:
  global flog, clog
  flog = _file_log(sys_log_path='{}/ck-py-log/'.format(work_path.get_user_work_path()), mixin=True, log_level=logging.DEBUG)
  clog = _console_log()
  GlobalLogger(flog)

def setLevel(log_level=logging.INFO):
  flog.setLevel(log_level)
  clog.setLevel(log_level)



import sched
import threading
import time
from functools import wraps

# annotation
def once(func):
  """
  decorator to initialize a function once
  :param func: function to be initialized
  :return: the real decorator for return the result
  """
  initialized = False
  res = None

  def wrapper(*args, **kwargs):
    nonlocal initialized, res
    if not initialized:
      res = func(*args, **kwargs)
      initialized = True
      return res
    else:
      return res

  return wrapper

# annotation
def init(func):
  """
  decorator to initialize a function automic
  :param func: function to be initialized
  :return: the real decorator for return the result
  """
  res = func()

  def wrapper():
    return res

  return wrapper

# annotation
def schd(interval_seconds, start_by_call: bool = False):
  start_flag = False
  run_flag = False
  scheduler = sched.scheduler(time.time, time.sleep)

  def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      nonlocal start_flag
      if start_by_call and not start_flag:
        start_flag = True
        threading.Thread(target=wrapper, args=args, kwargs=kwargs, daemon=True).start()
        return

      def job():
        func(*args, **kwargs)
        scheduler.enter(interval_seconds, 1, job)

      nonlocal run_flag
      if not run_flag:
        scheduler.enter(interval_seconds, 1, job)
        run_flag = True
        scheduler.run()
      else:
        func(*args, **kwargs)

    if not start_by_call: threading.Thread(target=wrapper, daemon=True).start()
    return wrapper

  return decorator



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2024/9/20 12:00'
import os
import time
from concurrent.futures import ProcessPoolExecutor

from ctools import call

_process_pool: ProcessPoolExecutor = None

@call.init
def init():
  global _process_pool
  max_workers = min(32, os.cpu_count()) # 最多 32 个
  _process_pool = ProcessPoolExecutor(max_workers=max_workers)

def cb(f, callback):
  exc = f.exception()
  if exc:
    if callback: callback(exc)
    raise exc
  else:
    if callback: callback(f.result())

def submit(func, *args, callback=None, **kwargs):
  if _process_pool is None: raise Exception('process pool is not init')
  future = _process_pool.submit(func, *args, **kwargs)
  future.add_done_callback(lambda f: cb(f, callback))
  time.sleep(0.01)
  return future

def shutdown(wait=True):
  if _process_pool is None: raise Exception('process pool is not init')
  _process_pool.shutdown(wait=wait)



import inspect
import os
import sys


def path(subPath: str = '') -> str:
  if getattr(sys, 'frozen', False):
    base_path = sys._MEIPASS
  else:
    caller_frame = inspect.currentframe().f_back
    caller_path = caller_frame.f_globals["__file__"]
    base_path = os.path.dirname(caller_path)
  return base_path + os.path.sep + subPath



import logging
import sys
import tkinter as tk


class Console:

  def __init__(self, master):
    self.master = master

    # 创建文本框和滚动条
    self.textbox = tk.Text(self.master, wrap=tk.NONE)

    self.vertical_scrollbar = tk.Scrollbar(self.textbox, command=self.textbox.yview)
    self.horizontal_scrollbar = tk.Scrollbar(self.textbox, command=self.textbox.xview, orient=tk.HORIZONTAL)

    self.textbox.configure(yscrollcommand=self.vertical_scrollbar.set, xscrollcommand=self.horizontal_scrollbar.set)
    self.textbox.pack(side=tk.LEFT, pady=10, padx=10, ipadx=10, ipady=10, fill=tk.BOTH, expand=True)

    self.vertical_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
    self.horizontal_scrollbar.pack(side=tk.BOTTOM, fill=tk.X)

    # 将标准输出和标准错误输出重定向到文本框中
    sys.stdout = self
    sys.stderr = self

    # # 创建输入框和按钮
    # self.entry = tk.Entry(self.master)
    # self.entry.pack(side=tk.BOTTOM, fill=tk.X, expand=True)
    # self.button = tk.Button(self.master, text="Send", command=self.send)
    # self.button.pack(side=tk.BOTTOM)

    # 将日志输出到文本框中
    self.log_handler = logging.StreamHandler(self)
    self.log_handler.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logging.getLogger().addHandler(self.log_handler)
    # logging.getLogger().setLevel(logging.INFO)

  def write(self, message):
    # 在文本框中输出消息
    self.textbox.insert(tk.END, message + '\n')
    self.textbox.see(tk.END)

  def flush(self):
    pass

  def send(self):
    # 获取输入框中的文本并打印到控制台
    text = self.entry.get()
    print(text)
    self.entry.delete(0, tk.END)

  def __del__(self):
    # 关闭日志处理器
    logging.getLogger().removeHandler(self.log_handler)



import tkinter
import tkinter as tk
from io import BytesIO
from tkinter import ttk

import requests
from PIL import Image, ImageTk

def showImageTip(root, title, imagePath, tips):
  # 创建一个Tk对象
  if root:
    window = root
  else:
    window = tk.Tk()
  # 设置窗口大小和位置
  win_width = 400
  win_height = 480
  screen_width = window.winfo_screenwidth()
  screen_height = window.winfo_screenheight()
  x = int((screen_width - win_width) / 2)
  y = int((screen_height - win_height) / 2)
  window.geometry("{}x{}+{}+{}".format(win_width, win_height, x, y))

  # 设置窗口大小和标题
  window.title(title)

  # 创建一个Label控件用于显示图片
  resp = requests.get(imagePath)
  image = Image.open(BytesIO(resp.content))  # 替换你自己的图片路径
  image = image.resize((400, 400))
  photo = ImageTk.PhotoImage(image)
  label1 = ttk.Label(window, image=photo)
  label1.pack(side=tkinter.TOP)

  # 创建一个Label控件用于显示提示文字
  label2 = ttk.Label(window, text=tips, font=("Arial Bold", 16))
  label2.config(anchor='center', justify='center')
  label2.pack(side=tkinter.BOTTOM)
  # 显示窗口
  window.mainloop()


if __name__ == '__main__':
  showImageTip(root=None, title='在线授权', imagePath='https://blog.gomyck.com/img/pay-img/wechatPay2Me.jpg', tips='{}\n授权已失效,请联系微信:\n{}'.format(123, 123))



import os
import shutil
import socket
import subprocess
import time
import zipfile
from configparser import ConfigParser

from PIL import Image

from ctools import work_path, call, sign, thread_pool
from ctools.sys_info import get_os_architecture, get_os_architecture4x

"""
本模块引用的依赖, 不允许引用sys_log模块, 否则将报错: 循环引用
打印使用print即可, 程序启动后, 会自动接管为log
"""


class Server:
  version = ''  # 版本信息
  port = 8010  # 服务端口
  metricsPort = 8011  # 获取指标信息端口
  wsPort = 8012  # websocket服务端口
  startTime = time.time()  # 系统启动时间
  baseWorkPath = work_path.get_user_work_path('rpa')  # 基础的工作目录
  sysLogPath = os.path.join(baseWorkPath, ".logs")  # 系统日志存储目录
  pythonHome = os.path.join(work_path.get_Users_path(), 'Public/python-3.8.9')  # python根目录
  indicatorsPath = os.path.join(baseWorkPath, "indicators")  # 指标信息存储目录
  screenshotPath = os.path.join(baseWorkPath, "screenshot")  # 截图存储目录
  controlServerAddr = None


class Authorization:
  enabled = True  # 访问权限验证开关
  whitelist = []


class Database:
  url_biz = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core", 'systemRDB')  # 数据库文件地址
  url_func = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core", 'explorerSearch')  # 函数库地址
  url_auth = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core", 'tmp')  # 授权库地址
  pool_size = 5
  max_overflow = 25


class Upload:
  upload_path = os.path.join(Server.baseWorkPath, 'uploads')  # 上传文件存储目录
  driver_path = os.path.join(Server.baseWorkPath, 'uploadDriver')  # 上传驱动存储目录
  source_pkg_path = os.path.join(Server.baseWorkPath, 'source_pkg')  # 资源中心位置
  module_path = os.path.join(Server.baseWorkPath, 'module_path')  # 上传组件存储目录


class Schedule:
  rpa_script_path = os.path.join(work_path.get_user_work_path(), ".cache/.tmp")  # 执行流程存储目录
  template_output_path = os.path.join(Server.baseWorkPath, 'document')  # 使用模板生成的文档存储目录
  play_wright_slow_mo = 0  # play wright每一步执行延时时间， 单位毫秒


@call.once
def sync_config():
  path = os.path.join(work_path.get_app_path(), "application.ini")
  conf = ConfigParser()
  if os.path.exists(path):
    conf.read(path)
    sections = conf.sections()
    ######### Server 必须放在第一个初始化的位置 ########
    if "Server" in sections:
      server_conf = conf['Server']
      Server.version = server_conf['version']
      Server.port = int(server_conf['port'])
      Server.baseWorkPath = work_path.get_user_work_path(server_conf['baseWorkPath'])
      Server.sysLogPath = os.path.join(Server.baseWorkPath, ".logs")
    ######### Server 必须放在第一个初始化的位置 ########
    if "Database" in sections:
      database = conf['Database']
      Database.url_biz = database['url_biz'] if database['url_biz'] else Database.url_biz
      Database.url_func = database['url_func'] if database['url_func'] else Database.url_func
      Database.url_auth = database['url_auth'] if database['url_auth'] else Database.url_auth
      Database.pool_size = int(database['poolSize'])
      Database.max_overflow = int(database['maxOverflow'])
    if "Upload" in sections:
      upload_path_section = conf['Upload']
      Upload.upload_path = os.path.join(Server.baseWorkPath, upload_path_section['path'])
      Upload.driver_path = os.path.join(Server.baseWorkPath, upload_path_section['driver_path'])
      Upload.source_pkg_path = os.path.join(Server.baseWorkPath, upload_path_section['source_pkg_path'])
    if "Schedule" in sections:
      schedule = conf['Schedule']
      Schedule.template_output_path = os.path.join(Server.baseWorkPath, schedule['templateOutputPath'])
      Schedule.play_wright_slow_mo = schedule['playWrightSlowMo']
    if "Authorization" in sections:
      auth = conf['Authorization']
      Authorization.enabled = bool(int(auth['enabled']))
      whitelist_paths = [w.strip() for w in auth['whitelist'].split(',')]
      Authorization.whitelist.extend(whitelist_paths)

    os.makedirs(Server.baseWorkPath, exist_ok=True)
    os.makedirs(Server.indicatorsPath, exist_ok=True)
    os.makedirs(Server.screenshotPath, exist_ok=True)
    os.makedirs(Upload.source_pkg_path, exist_ok=True)
    os.makedirs(Schedule.template_output_path, exist_ok=True)
    os.makedirs(Upload.module_path, exist_ok=True)
  else:
    print('配置文件不存在!')
    raise Exception('配置文件不存在!')


def check_database():
  db_file = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core", 'systemRDB')
  if os.path.exists(db_file):
    db_size = os.path.getsize(db_file)
    if db_size < 2000 * 1024: os.remove(db_file)
  db_file = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core", 'tmp')
  if os.path.exists(db_file):
    db_size = os.path.getsize(db_file)
    if db_size < 2000 * 1024: os.remove(db_file)

def validate_sign():
  sign_app_val = sign.generate_signature(work_path.get_install_path('rpa-server.exe'))
  with open(work_path.get_install_path('rpa-server.sign')) as sign_file:
    sign_file_val = sign_file.readline()
  if 'rpa-dev-sign' == sign_file_val.strip(): return
  if sign_app_val != sign_file_val: raise Exception('程序签名验证失败!!!')


def path_config():
  print('start config path')
  driverPath = Upload.driver_path
  pythonPath = os.path.join(work_path.get_Users_path(), 'Public/python-3.8.9')
  hplPath = os.path.join(work_path.get_user_work_path(), 'AppData/Local/ms-playwright')
  taguiPath = os.path.join(work_path.get_user_work_path(), 'AppData/Roaming/tagui')
  os.environ['PATH'] = os.pathsep.join([pythonPath, os.path.join(pythonPath, 'Scripts'), driverPath, os.path.join(hplPath, 'chromium-920619\chrome-win'), taguiPath, os.environ['PATH']])
  print('end config path')


def sync_version(callFunc):
  destFilePath = os.path.join(Server.baseWorkPath, "version")
  driverPath = Upload.driver_path
  pythonPath = os.path.join(work_path.get_Users_path(), 'Public/python-3.8.9')
  msPlayPath = os.path.join(work_path.get_user_work_path(), 'AppData/Local/ms-playwright')
  taguiPath = os.path.join(work_path.get_user_work_path(), 'AppData/Roaming/tagui')
  if not os.path.exists(destFilePath):
    try:
      shutil.rmtree(pythonPath)
    except Exception:
      pass
    try:
      shutil.rmtree(msPlayPath)
    except Exception:
      pass
    try:
      shutil.rmtree(driverPath)
    except Exception:
      pass
    try:
      shutil.rmtree(taguiPath)
    except Exception:
      pass
    from ctools.pacth import Patch
    patch = Patch(oldVersion='V1.0.0', newVersion=Server.version, pythonPath=pythonPath, playwrightPath=msPlayPath, driverPath=driverPath)
    patch.apply_patch()
    if callFunc: callFunc()
    with open(destFilePath, 'w') as newVersion:
      newVersion.write(Server.version)
      print('初始化安装, 版本信息为: {}'.format(Server.version))
      newVersion.flush()
  else:
    with open(destFilePath, 'r') as oldVersion:
      oldV = oldVersion.readline()
      print('本地版本信息为: {}, 程序版本信息为: {}'.format(oldV, Server.version))
      oldVersion.close()
    if oldV == Server.version and '-snapshot' not in oldV: return
    print('开始升级本地程序..')
    from ctools.pacth import Patch
    patch = Patch(oldVersion=oldV, newVersion=Server.version, pythonPath=pythonPath, playwrightPath=msPlayPath, driverPath=driverPath)
    patch.apply_patch()
    if callFunc: callFunc()
    with open(destFilePath, 'w') as newVersion:
      newVersion.write(Server.version)
      print('程序升级成功, 更新版本信息为: {}'.format(Server.version))
      newVersion.flush()


def sync_database():
  for db in [['simpleYWI', 'systemRDB', False], ['simpleHSI', 'explorerSearch', True], ['simpleSQI', 'tmp', False]]:
    dbPath = os.path.join(work_path.get_app_path(), "extension", db[0])
    destDBPath = os.path.join(work_path.get_user_work_path(), "AppData/Local/SystemWin32Core")
    os.makedirs(destDBPath, exist_ok=True)
    destFilePath = os.path.join(destDBPath, db[1])
    try:
      if db[2] and os.path.exists(destFilePath): os.remove(destFilePath)
    except Exception:
      pass
    if not os.path.exists(destFilePath):
      shutil.copy(dbPath, destFilePath)
    else:
      pass


def sync_python_env():
  print('SDK 开始安装...')
  sourcePath = os.path.join(work_path.get_install_path(), "sources", 'python-3.8.9-{}.zip'.format(get_os_architecture()))
  destPLPath = os.path.join(work_path.get_Users_path(), 'Public/python-3.8.9')
  unzipPLPath = os.path.join(work_path.get_Users_path(), 'Public')
  if not os.path.exists(destPLPath):
    with zipfile.ZipFile(sourcePath, 'r') as zip_ref:
      zip_ref.extractall(unzipPLPath)
    print('SDK 安装成功(0)')
  else:
    print('SDK 安装成功(1)')
  print('SDK 环境变量配置成功')


def sync_python_module():
  print('SDK 安装模块开始')
  module_list = []

  if os.path.exists(Upload.module_path):
    for file_name in os.listdir(Upload.module_path):
      m_path = os.path.join(Upload.module_path, file_name)
      module_list.append(m_path)

  upgrade_module_path = os.path.join(work_path.get_install_path(), "sources/UpgradeModule")
  init_flag = os.path.join(upgrade_module_path, "init")
  upgrade_module_path = os.path.join(upgrade_module_path, get_os_architecture4x())
  if not os.path.exists(init_flag) and os.path.exists(upgrade_module_path):
    for file_name in os.listdir(upgrade_module_path):
      m_path = os.path.join(upgrade_module_path, file_name)
      module_list.append(m_path)

  if module_list:

    freeze_msg = ""
    pip_path = os.path.join(Server.pythonHome, 'Scripts/pip.exe')
    cmd = [pip_path, "freeze"]
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    if proc:
      stdout, stderr = proc.communicate()
      freeze_msg = str(stdout)

    for m_path in module_list:
      if os.path.isfile(m_path):
        installed = False
        m = os.path.basename(m_path)
        m_name = m.split("-")
        if len(m_name) >= 2:
          installed = "%s-%s" % (m_name[0], m_name[1]) in freeze_msg or "%s==%s" % (m_name[0], m_name[1]) in freeze_msg

        if not installed:
          cmd = [pip_path, "install", "--no-dependencies", m_path]
          proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE, close_fds=True)
          if proc:
            stdout, stderr = proc.communicate()
            if stdout:
              if "Successfully installed" in str(stdout) or "already installed" in str(stdout):
                print("SDK 安装本地模块%s成功" % m)
              else:
                print("SDK 安装本地模块%s失败" % m)

    if not os.path.exists(init_flag):
      with open(init_flag, "w") as f:
        f.write("success")
        f.flush()
  print('SDK 安装本地模块完成')


def sync_python_interpreter():
  destPLPath = os.path.join(work_path.get_Users_path(), 'Public/python-3.8.9')
  enhancePath = os.path.join(work_path.get_Users_path(), "Public/enhance", 'enhance-{}.zip'.format(get_os_architecture()))
  originPath = os.path.join(work_path.get_Users_path(), "Public/enhance", 'origin-{}.zip'.format(get_os_architecture()))
  initFlag = os.path.join(work_path.get_Users_path(), "Public/enhance", 'init')
  flagValue = 'can not find dest dir'
  os.makedirs(os.path.join(work_path.get_Users_path(), "Public/enhance"), exist_ok=True)
  if not os.path.exists(initFlag):
    try:
      if os.path.exists(enhancePath):
        with zipfile.ZipFile(enhancePath, 'r') as zip_ref:
          zip_ref.extractall(destPLPath)
          flagValue = 'enhance'
        print('enhance enabled..')
      elif os.path.exists(originPath):
        with zipfile.ZipFile(originPath, 'r') as zip_ref:
          zip_ref.extractall(destPLPath)
          flagValue = 'origin'
        print('origin enabled..')
      with open(initFlag, 'w') as flag:
        flag.write(flagValue)
        print('interpreter init success: {}'.format(flagValue))
        flag.flush()
    except Exception as e:
      print('interpreter init error: {}'.format(e))
  else:
    with open(initFlag, 'r') as flag:
      flagValue = flag.readline()
      print('interpreter flag: {}'.format(flagValue))
      flag.close()


def sync_driver():
  print('DRIVER 开始安装...')
  sourcePath = os.path.join(work_path.get_install_path(), "sources", 'uploadDriver.zip')
  destPLPath = Upload.driver_path
  unzipPLPath = Upload.driver_path
  if not os.path.exists(os.path.join(destPLPath)):
    with zipfile.ZipFile(sourcePath, 'r') as zip_ref:
      zip_ref.extractall(unzipPLPath)
    print('DRIVER 安装成功(0)')
  else:
    print('DRIVER 安装成功(1)')

  print('DRIVER 配置成功')


def sync_tagui():
  print('HTG ENGINE 开始安装...')
  plPath = os.path.join(work_path.get_install_path(), "sources", 'htg-source.zip')
  destPLPath = os.path.join(work_path.get_user_work_path(), 'AppData/Roaming/tagui')
  unzipPLPath = os.path.join(work_path.get_user_work_path(), 'AppData/Roaming/')
  if not os.path.exists(os.path.join(destPLPath)):
    with zipfile.ZipFile(plPath, 'r') as zip_ref:
      zip_ref.extractall(unzipPLPath)
    print('HTG ENGINE 依赖安装成功(0)')
  else:
    print('HTG ENGINE 依赖安装成功(1)')


def sync_playwright():
  print('HPL ENGINE 开始安装...')
  plPath = os.path.join(work_path.get_install_path(), "sources", 'ms-playwright.zip')
  destPLPath = os.path.join(work_path.get_user_work_path(), 'AppData/Local/ms-playwright')
  unzipPLPath = os.path.join(work_path.get_user_work_path(), 'AppData/Local/')
  if not os.path.exists(os.path.join(destPLPath)):
    with zipfile.ZipFile(plPath, 'r') as zip_ref:
      zip_ref.extractall(unzipPLPath)
    print('HPL ENGINE 依赖安装成功(0)')
  else:
    print('HPL ENGINE 依赖安装成功(1)')

def sync_paddleocr_model():
  model_path = os.path.join(work_path.get_install_path(), "sources", 'paddleocr-model.zip')
  paddleocr_path = work_path.get_user_work_path(".paddleocr")
  if not os.path.exists(paddleocr_path) and os.path.exists(model_path) and get_os_architecture() == "64":
    print('安装paddleocr模型文件开始')
    with zipfile.ZipFile(model_path, 'r') as zip_ref:
      zip_ref.extractall(paddleocr_path)
    print('安装paddleocr模型文件结束')


def clean_temp_dir(target_file='rpa_ident'):
  root_folder = work_path.get_user_temp_path()
  print("Start clear cache...")
  for folder_name in os.listdir(root_folder):
    folder_path = os.path.join(root_folder, folder_name)
    if os.path.isdir(folder_path):
      file_path = os.path.join(folder_path, target_file)
      if os.path.exists(file_path) and os.path.normpath(folder_path) != os.path.normpath(work_path.get_app_path()):
        try:
          shutil.rmtree(folder_path)
        except Exception:
          pass
  print("End clear cache...")
  shutil.rmtree(Schedule.rpa_script_path, ignore_errors=True)


def is_port_in_use(port):
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    return s.connect_ex(('localhost', port)) == 0


def system_tray(app_server):
  pystray = __import__('pystray')
  try:
    def on_quit_clicked(app_icon):
      thread_pool.submit(app_server.stop)
      app_icon.stop()

    image = Image.open(os.path.join(work_path.get_app_path(), "images/ico.png"))
    menu = (
      pystray.MenuItem(text='退出', action=on_quit_clicked),
    )
    icon = pystray.Icon("name", image, "RPA客户端", menu)
    icon.run()
  except Exception:
    pass



import _thread
import queue
import time

# 伪终端交互

from winpty import PtyProcess


class Code:
  SUCCESS = 200
  FAIL = 201
  PART = 206


class PtyTools:
  def __init__(self, cmd, prefix: str = ""):
    self._process = PtyProcess.spawn(cmd)
    self._read_queue = queue.Queue()
    self._prefix = prefix
    self._is_enable = False

  def _read(self):
    try:
      while True:
        msg = self._process.readline()
        self._is_enable = True
        if msg and self._prefix not in msg[:len(self._prefix)]:
          self._read_queue.put(msg)
        elif self._is_exit():
          break
    except Exception:
      pass

  def _is_exit(self):
    return self._process.exitstatus is not None

  def run(self):
    _thread.start_new_thread(self._read, ())
    while not self._is_enable:
      time.sleep(1)

  def read(self, end_str: str = None, is_async=False, timeout: int = 1):
    code = Code.SUCCESS
    content = ""
    start_time = time.time()
    while True:
      if not self._read_queue.empty():
        line = self._read_queue.get()
        content += line
        start_time = time.time()
        if end_str is not None and end_str in line:
          break
        if is_async:
          code = Code.PART
          break
      else:
        if time.time() - start_time >= timeout:
          break
        if self._is_exit():
          code = Code.FAIL
          break
      time.sleep(0.1)
    return code, content

  def send(self, msg, timeout: int = 0.5):
    self._process.write('%s\r\n' % msg)
    return self.read(timeout=timeout)

  def close(self):
    self._process.close(force=True)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2025/1/24 08:48'

import io
import time

import pyzipper

def add_file_to_zip(file_name, file_bytes:[], password=None) -> io.BytesIO:
  zip_filename = "{}}_{}.zip".format(file_name, time.strftime('%Y-%m-%d_%H-%M-%S-%s', time.localtime(time.time())))
  zipFile = io.BytesIO()
  with pyzipper.AESZipFile(zipFile, 'w', compression=pyzipper.ZIP_DEFLATED, encryption=pyzipper.WZ_AES) as zipf:
    if password: zipf.setpassword(password.encode('utf-8'))
    for file in file_bytes:
      zipf.writestr(zip_filename, file)
  zipFile.seek(0)
  return zipFile



from openpyxl import load_workbook
from openpyxl.worksheet.datavalidation import DataValidation


class excelUtil:
  wb = None
  sourcePath = None
  savePath = None

  def __init__(self, path, save_path):
    # 创建一个 Workbook 对象
    self.wb = load_workbook(path)
    # 在 Workbook 中创建一个 Worksheet 对象
    self.ws = self.wb.active
    self.sourcePath = path
    self.savePath = save_path

  def makeDropData(self, col, drop_data):
    # 定义下拉框的数据
    dropdown_items = drop_data
    # 将下拉框数据转换成字符串
    dropdown_items_str = ','.join(dropdown_items)
    # 在第一列中添加下拉框
    dropdown_col = col
    dropdown_start_row = 2
    dropdown_end_row = 200
    # 配置下拉框参数
    dropdown = DataValidation(type="list", formula1=f'"{dropdown_items_str}"', allow_blank=True)
    # 添加下拉框到指定的单元格区域
    dropdown_range = f"{dropdown_col}{dropdown_start_row}:{dropdown_col}{dropdown_end_row}"
    self.ws.add_data_validation(dropdown)
    dropdown.add(dropdown_range)

  def save(self):
    # 保存工作簿
    self.wb.save(self.savePath)



import contextlib
import datetime
import math

from sqlalchemy import create_engine, Integer, Column, event
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.sql import text

from ctools import call, string_tools
from ctools.thread_pool import thread_local

"""
class XXXX(BaseMixin):
  __tablename__ = 't_xxx_info'
  __table_args__ = {'comment': 'xxx信息表'}
  server_content: Column = Column(String(50), nullable=True, default='', comment='123123')
  server_ip: Column = Column(String(30), index=True)
  user_id: Column = Column(BigInteger)

database.init_db('postgresql://postgres:123456@192.168.3.107:32566/abc', default_schema='public', db_key='source', pool_size=100)
with database.get_session('source') as s:
  s.execute(text('insert into xxx (name) values (:name)'), {'name': string_tools.get_random_str(5)})
  s.commit()
"""

Base = None
inited_db = {}
engines = {}
sessionMakers = {}

def getEngine(db_key: str='default'):
  return engines[db_key]

@call.init
def _init():
  global Base
  Base = declarative_base()

"""
The string form of the URL is
dialect[+driver]://user:password@host/dbname[?key=value..]
where ``dialect`` is a database name such as ``mysql``, ``oracle``, ``postgresql``, etc.
and ``driver`` the name of a DBAPI such as ``psycopg2``, ``pyodbc``, ``cx_oracle``, etc.  Alternatively
"""

# 密码里的@  要替换成 %40

# sqlite connect_args={"check_same_thread": False}  db_url=sqlite:///{}.format(db_url)
# sqlite 数据库, 初始化之后, 优化一下配置
# $ sqlite3 app.db
# > PRAGMA journal_mode=WAL; 设置事务的模式, wal 允许读写并发, 但是会额外创建俩文件
# > PRAGMA synchronous=NORMAL; 设置写盘策略, 默认是 FULL, 日志,数据都落, 设置成 NORMAL, 日志写完就算事务完成

def init_db(db_url: str, db_key: str='default', connect_args: dict={}, default_schema: str=None, pool_size: int=5, max_overflow: int=25, echo: bool=False):
  if db_url.startswith('mysql'):
    import pymysql
    pymysql.install_as_MySQLdb()
  if inited_db.get(db_key): raise Exception('db {} already init!!!'.format(db_key))
  global engines, sessionMakers
  engine, sessionMaker = _create_connection(db_url=db_url, connect_args=connect_args, pool_size=pool_size, max_overflow=max_overflow, echo=echo)
  engines[db_key] = engine
  sessionMakers[db_key] = sessionMaker
  inited_db[db_key] = True
  if default_schema: event.listen(engine, 'connect', lambda dbapi_connection, connection_record: _set_search_path(dbapi_connection, default_schema))
  Base.metadata.create_all(engine)

def _set_search_path(dbapi_connection, default_schema):
  with dbapi_connection.cursor() as cursor:
    cursor.execute(f'SET search_path TO {default_schema}')

def _create_connection(db_url: str, pool_size: int=5, max_overflow: int=25, connect_args={}, echo: bool=False):
  engine = create_engine('{}'.format(db_url),
                               echo=echo,
                               future=True,
                               pool_size=pool_size,
                               max_overflow=max_overflow,
                               pool_pre_ping=True,
                               pool_recycle=3600,
                               connect_args=connect_args)
  sm = sessionmaker(bind=engine)
  return engine, sm

def generate_custom_id():
  return str(string_tools.get_snowflake_id())

class BaseMixin(Base):
  __abstract__ = True
  obj_id = Column(Integer, primary_key=True, default=generate_custom_id)
  # ext1 = Column(String)
  # ext2 = Column(String)
  # ext3 = Column(String)
  # create_time = Column(DateTime, nullable=False, default=datetime.datetime.now)
  # update_time = Column(DateTime, nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now, index=True)

  def to_dict(self):
    return self.__getstate__()

  def from_dict(self, v):
    self.__dict__.update(v)

  def __getstate__(self):
    ret_state = {}
    state = self.__dict__.copy()
    for key in state.keys():
      if not key.startswith("_"):
        if type(state[key]) == datetime.datetime:
          ret_state[key] = state[key].strftime("%Y-%m-%d %H:%M:%S")
        else:
          ret_state[key] = state[key]
    return ret_state

@contextlib.contextmanager
def get_session(db_key: str='default') -> Session:
  thread_local.db_key = db_key
  if sm:=sessionMakers.get(db_key):
    s = sm()
  else:
    raise ValueError("Invalid db_key: {}".format(db_key))
  try:
    yield s
  except Exception as e:
    s.rollback()
    raise e
  finally:
    s.close()

class PageInfoBuilder:

  def __init__(self, pageInfo, total_count, records):
    self.page_size = pageInfo.page_size
    self.page_index = pageInfo.page_index
    self.total_count = total_count
    self.total_page = math.ceil(total_count / int(pageInfo.page_size))
    self.records = records

def query_by_page(query, pageInfo):
  records = query.offset((pageInfo.page_index - 1) * pageInfo.page_size).limit(pageInfo.page_size).all()
  rs = []
  for r in records:
    rs.append(r)
  return PageInfoBuilder(pageInfo, query.count(), rs)

def query4_crd_sql(session, sql: str, params: dict) -> []:
  records = session.execute(text(sql), params).fetchall()
  rs = []
  for record in records:
    data = {}
    for index, key in enumerate(record._mapping):
      data[key] = record[index]
    rs.append(data)
  return rs

sqlite_and_pg_page_sql = """
 limit :limit offset :offset
"""
mysql_page_sql = """
 limit :offset, :limit
"""

def query_by_page4_crd_sql(session, sql: str, params: dict, pageInfo) -> []:
  db_name = engines[thread_local.db_key].name
  if db_name == 'postgresql' or db_name == 'sqlite':
    page_sql = sqlite_and_pg_page_sql
  elif db_name == 'mysql':
    page_sql = mysql_page_sql
  else:
    raise Exception('not support db: {}'.format(db_name))
  wrapper_sql = """
    select * from ({}) as t {}
  """.format(sql, page_sql)
  count_sql = """
    select count(1) from ({}) as t
  """.format(sql)
  params["limit"] = pageInfo.page_size
  params["offset"] = (pageInfo.page_index - 1) * pageInfo.page_size
  records = session.execute(text(wrapper_sql), params).fetchall()
  rs = []
  for record in records:
    data = {}
    for index, key in enumerate(record._mapping):
      data[key] = record[index]
    rs.append(data)
  return PageInfoBuilder(pageInfo, session.execute(text(count_sql), params).first()[0], rs)



import inspect
import threading
from functools import wraps

import bottle
from bottle import response, Bottle, request

from ctools import ctoken
from ctools.dict_wrapper import DictWrapper
from ctools.api_result import R
from ctools.sys_log import flog as log

bottle.BaseRequest.MEMFILE_MAX = 1024 * 1024 * 50
func_has_params = {}

class GlobalState:
  lock = threading.Lock()
  withOutLoginURI = [
    '/',
    '/index',
    '/login'
  ]
  allowRemoteCallURI = [

  ]
  token = {}
  interceptors = []

def init_app(context_path=None):
  app = Bottle()
  app.context_path = context_path

  @app.hook('before_request')
  def before_request():
    for interceptor in GlobalState.interceptors:
      res: R = interceptor['func']()
      if res.code != 200: bottle.abort(res.code, res.message)

  @app.error(401)
  def unauthorized(error):
    response.status = 401
    log.error("系统未授权: {} {} {}".format(error.body, request.method, request.fullpath))
    return R.error(resp=R.Code.cus_code(401, "系统未授权! {}".format(error.body)))

  @app.error(403)
  def unauthorized(error):
    response.status = 403
    log.error("访问受限: {} {} {}".format(error.body, request.method, request.fullpath))
    return R.error(resp=R.Code.cus_code(403, "访问受限: {}".format(error.body)))

  @app.error(404)
  def not_found(error):
    response.status = 404
    log.error("404 not found : {} {} {}".format(error.body, request.method, request.fullpath))
    return R.error(resp=R.Code.cus_code(404, "资源未找到: {}".format(error.body)))

  @app.error(405)
  def method_not_allow(error):
    response.status = 405
    log.error("请求方法错误: {} {} {}".format(error.status_line, request.method, request.fullpath))
    return R.error(resp=R.Code.cus_code(405, '请求方法错误: {}'.format(error.status_line)))

  @app.error(500)
  def internal_error(error):
    response.status = 500
    log.error("系统发生错误: {} {} {}".format(error.body, request.method, request.fullpath))
    return R.error(msg='系统发生错误: {}'.format(error.exception))

  @app.hook('after_request')
  def after_request():
    enable_cors()

  app.install(params_resolve)
  return app

def enable_cors():
  response.headers['Access-Control-Allow-Origin'] = '*'
  response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, DELETE, OPTIONS'
  request_headers = request.headers.get('Access-Control-Request-Headers')
  response.headers['Access-Control-Allow-Headers'] = request_headers if request_headers else ''
  response.headers['Access-Control-Expose-Headers'] = '*'

# annotation
def before_intercept(order=0):
  def decorator(func):
    log.info("add before interceptor: {}".format(func.__name__))
    GlobalState.interceptors.append({'order': order, 'func': func})
    GlobalState.interceptors = sorted(GlobalState.interceptors, key=lambda x: x['order'])
  return decorator

# annotation
def rule(key):
  def return_func(func):
    @wraps(func)
    def decorated(*args, **kwargs):
      # if GlobalState.licenseInfo is not None and key not in GlobalState.licenseInfo.access_module:
      #   log.error("系统未授权! {} {}".format(request.fullpath, '当前请求的模块未授权!请联系管理员!'))
      #   return R.error(resp=R.Code.cus_code(9999, "系统未授权! {}".format('当前请求的模块未授权!请联系管理员!')))
      return func(*args, **kwargs)
    return decorated
  return return_func

# annotation or plugins, has auto install, don't need to call
def params_resolve(func):
  @wraps(func)
  def decorated(*args, **kwargs):
    if func_has_params.get(request.fullpath) is not None and not func_has_params.get(request.fullpath):
      return func(*args, **kwargs)
    if func_has_params.get(request.fullpath) is None:
      sig = inspect.signature(func)
      params = sig.parameters
      if not params.get('params'):
        func_has_params[request.fullpath] = False
        return func(*args, **kwargs)
      else:
        func_has_params[request.fullpath] = True
    if request.method == 'GET':
      queryStr = request.query.decode('utf-8')
      page_info = PageInfo(
        page_size=10 if request.headers.get('page_size') is None else int(request.headers.get('page_size')),
        page_index=1 if request.headers.get('page_index') is None else int(request.headers.get('page_index'))
      )
      queryStr.page_info = page_info
      return func(params=queryStr, *args, **kwargs)
    elif request.method == 'POST':
      query_params = request.query.decode('utf-8')
      content_type = request.get_header('content-type')
      if content_type == 'application/json':
        params = request.json or {}
        dict_wrapper = DictWrapper(params)
        dict_wrapper.query_params = query_params
        return func(params=dict_wrapper, *args, **kwargs)
      elif 'multipart/form-data' in content_type:
        form_data = request.forms.decode()
        form_files = request.files.decode()
        dict_wrapper = DictWrapper(form_data)
        dict_wrapper.query_params = query_params
        dict_wrapper.files = form_files
        return func(params=dict_wrapper, *args, **kwargs)
      elif 'application/x-www-form-urlencoded' in content_type:
        params = request.forms.decode()
        dict_wrapper = DictWrapper(params.dict)
        dict_wrapper.query_params = query_params
        return func(params=dict_wrapper, *args, **kwargs)
      elif 'text/plain' in content_type:
        params = request.body.read().decode('utf-8')
        dict_wrapper = DictWrapper({'body': params})
        dict_wrapper.query_params = query_params
        return func(params=dict_wrapper, *args, **kwargs)
    else:
      return func(*args, **kwargs)
  return decorated

class PageInfo:
  def __init__(self, page_size, page_index):
    self.page_size = page_size
    self.page_index = page_index

# 通用的鉴权方法
def common_auth_verify(aes_key):
  if request.path.startswith('/static') or request.path in GlobalState.withOutLoginURI:
    return R.ok(to_json_str=False)
  auth_token = request.get_header('Authorization')
  if auth_token is None:
    auth_token = request.get_cookie('Authorization')
  payload = ctoken.get_payload(auth_token, aes_key)
  if payload:
    return R.ok(to_json_str=False)
  return R.error(resp=R.Code.cus_code(401, "未授权"), to_json_str=False)



import time
from datetime import datetime, timedelta

def get_date():
  return time.strftime('%Y-%m-%d', time.localtime(time.time()))

def get_time():
  return time.strftime('%H-%M-%S', time.localtime(time.time()))

def get_date_time(fmt="%Y-%m-%d %H:%M:%S"):
  return time.strftime(fmt, time.localtime(time.time()))

def str_to_datetime(val: str, fmt="%Y-%m-%d %H:%M:%S"):
  return time.strptime(val, fmt)

def str_to_timestamp(val: str, fmt="%Y-%m-%d %H:%M:%S"):
  return time.mktime(time.strptime(val, fmt))

def timestamp_to_str(timestamp: int=time.time(), fmt="%Y-%m-%d %H:%M:%S"):
  return time.strftime(fmt, time.localtime(timestamp))

def get_today_start_end(now: datetime.now()):
  start = datetime(now.year, now.month, now.day, 0, 0, 0)
  end = datetime(now.year, now.month, now.day, 23, 59, 59)
  return start.strftime("%Y-%m-%d %H:%M:%S"), end.strftime("%Y-%m-%d %H:%M:%S")

def get_week_start_end(now: datetime.now()):
  start = now - timedelta(days=now.weekday())  # 本周一
  end = start + timedelta(days=6)  # 本周日
  return start.strftime("%Y-%m-%d 00:00:00"), end.strftime("%Y-%m-%d 23:59:59")

def time_diff_in_seconds(sub_head: str=get_date_time(), sub_end: str=get_date_time()):
  start_ts = str_to_timestamp(sub_head)
  end_ts = str_to_timestamp(sub_end)
  return int(start_ts - end_ts)

def opt_time(base_time=None, days=0, hours=0, minutes=0, seconds=0, weeks=0, fmt="%Y-%m-%d %H:%M:%S"):
  if base_time is None:
    base_time = datetime.now()
  elif isinstance(base_time, str):
    base_time = datetime.strptime(base_time, fmt)
  new_time = base_time + timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds, weeks=weeks)
  return new_time.strftime(fmt)



from bs4 import BeautifulSoup

from ctools.ex import exception_handler


@exception_handler(fail_return=['解析错误'], print_exc=True)
def table2list(html, include_header=True, recursive_find=True,
               table_tag='table', table_class=None, table_attrs: dict = {},
               row_tag='tr', row_class=None, row_attrs: dict = {},
               header_cell_tag='th', header_cell_class=None, header_cell_attrs: dict = {},
               cell_tag='td', cell_class=None, cell_attrs: dict = {}):
  soup = BeautifulSoup(markup=html, features='html.parser')
  if table_class:
    table = soup.find(table_tag, class_=table_class, **table_attrs)
  else:
    table = soup.find(table_tag, **table_attrs)
  if row_class:
    all_row = table.find_all(row_tag, class_=row_class, recursive=recursive_find, **row_attrs)
  else:
    all_row = table.find_all(row_tag, recursive=recursive_find, **row_attrs)
  rows = []
  if include_header:
    if header_cell_class:
      header = [i.text for i in all_row[0].find_all(header_cell_tag, class_=header_cell_class, recursive=recursive_find, **header_cell_attrs)]
    else:
      header = [i.text for i in all_row[0].find_all(header_cell_tag, recursive=recursive_find, **header_cell_attrs)]
    rows.append(header)
  for tr in all_row[1 if include_header else 0:]:
    if cell_class:
      td = tr.find_all(cell_tag, class_=cell_class, recursive=recursive_find, **cell_attrs)
    else:
      td = tr.find_all(cell_tag, recursive=recursive_find, **cell_attrs)
    row = [i.text for i in td]
    rows.append(row)
  return rows



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2025/2/14 11:09'

import redis
from redis import Redis

from ctools import date_utils, thread_pool, string_tools

def init_pool(host: str = 'localhost', port: int = 6379, db: int = 0, password: str = None,
              username: str = None, decode_responses: bool = True, max_connections: int = 75,
              health_check_interval: int = 30, retry_count: int = 3) -> Redis:
    for attempt in range(retry_count):
        try:
            r: Redis = redis.StrictRedis(
                host=host, port=port, db=db,
                username=username, password=password,
                retry_on_timeout=True,
                max_connections=max_connections,
                decode_responses=decode_responses,
                health_check_interval=health_check_interval,
                socket_connect_timeout=5,
                socket_timeout=5
            )
            if r.ping():
                print('CRedis connect {} {} success!'.format(host, port))
                return r
        except redis.ConnectionError as e:
            if attempt == retry_count - 1:
                raise Exception(f"Failed to connect to Redis after {retry_count} attempts: {str(e)}")
            print(f"Connection attempt {attempt + 1} failed, retrying...")

def add_lock(r: Redis, key: str, timeout: int = 30):
  if r.exists(key):
    expire_time = r.get(key)
    if date_utils.time_diff_in_seconds(expire_time, date_utils.get_date_time()) > 0:
      return True
    else:
      r.delete(key)
  return r.set(key, date_utils.opt_time(seconds=timeout), nx=True, ex=timeout) is not None

def remove_lock(r: Redis, key: str):
  r.delete(key)

def subscribe(r: Redis, channel_name, callback):
  def thread_func():
    pubsub = r.pubsub()
    pubsub.subscribe(channel_name)
    for message in pubsub.listen():
      callback(message)
  thread_pool.submit(thread_func)

def _process_pending_messages(r: Redis, stream_name: str, group_name: str, consumer_name: str, callback):
  """
  处理未确认的消息
  :param r: Redis 连接
  :param stream_name: 流名称
  :param group_name: 消费者组名称
  :param consumer_name: 消费者名称
  :param callback: 消息处理回调函数
  """
  # 检查未确认的消息
  pending_messages = r.xpending(stream_name, group_name)
  if pending_messages['pending'] > 0:
    print(f"Found {pending_messages['pending']} pending messages.")
    # 获取未确认的消息列表
    pending_list = r.xpending_range(stream_name, group_name, min='-', max='+', count=pending_messages['pending'])
    for message in pending_list:
      message_id = message['message_id']
      claimed_messages = r.xclaim(stream_name, group_name, consumer_name, min_idle_time=0, message_ids=[message_id])
      if claimed_messages:
        # 处理消息
        for claimed_message in claimed_messages:
          message_id, data = claimed_message
          print(f"Processing pending message: {message_id}, data: {data}")
          try:
            if callback(message_id, data):
              r.xack(stream_name, group_name, message_id)
          except Exception as e:
            print(f"Error processing message {message_id}: {e}")
  else:
    print("No pending messages found.")

def stream_subscribe(r: Redis, stream_name, group_name, callback, from_id: str='$', noack: bool = False):
  def thread_func():
    try:
      # $表示从最后面消费, 0表示从开始消费
      r.xgroup_create(name=stream_name, groupname=group_name, id=from_id, mkstream=True)
      print(f"Consumer group '{group_name}' created successfully.")
    except Exception as e:
      if "already exists" in str(e):
        print(f"Consumer group '{group_name}' already exists.")
      else:
        print(f"Error creating consumer group '{group_name}': {e}")
    consumer_name = 'consumer-{}'.format(string_tools.get_uuid())
    # 处理未确认的消息
    _process_pending_messages(r, stream_name, group_name, consumer_name, callback)
    while True:
      messages = r.xreadgroup(group_name, consumer_name, {stream_name: '>'}, block=1000, noack=noack)
      for message in messages:
        try:
          message_id, data = message[1][0]
          res = callback(message_id, data)
          if res: r.xack(stream_name, group_name, message_id)
        except Exception as e:
          print('stream_subscribe error: ', e)
  thread_pool.submit(thread_func)

def stream_publish(r: Redis, stream_name, message):
  r.xadd(stream_name, message)



import time
from enum import Enum
from typing import Dict

from ctools.dict_wrapper import DictWrapper as DictToObj
from paho.mqtt import client as mqtt
from paho.mqtt.enums import CallbackAPIVersion

from ctools import sys_log, cjson, string_tools, sys_info, date_utils, sm_tools, thread_pool

class MQTTEvent(Enum):

  pass

'''
MQTT服务使用示例:

mqtt_utils = MQTTUtils("192.168.3.100")
mqtt_utils.connect(username="rpa", password="123")

# 订阅
mqtt_utils.subscribe("test/topic1", "test/topic2")

# 异步发送
msg = {"name": "测试消息"}
mqtt_utils.publish_async(msg, event=MQTTEvent.END_TASK)

# 同步发送
mqtt_resp = mqtt_utils.publish_sync(msg, event=MQTTEvent.UPLOAD_PROCESS, timeout=60)
status = mqtt_resp.status
body = mqtt_resp.body
print("%s, %s" % (status, body))
'''

log = sys_log.flog

NO_LOG_EVENT = [MQTTEvent.HEARTBEAT]


class MQTTRequest:

  def __init__(self, **kwargs):
    self.event: str = kwargs.get('event')
    self.trace_id: str = kwargs.get('trace_id') if kwargs.get('trace_id') else string_tools.get_uuid()
    self.client_id: str = kwargs.get('client_id')
    self.time: str = kwargs.get('time') if kwargs.get('time') else date_utils.get_date_time()
    self.body = kwargs.get('body')

  def __str__(self):
    res_str = ""
    for k, v in self.__dict__.items():
      res_str += "%s=%s," % (k, v)
    return res_str[:-1]


class MQTTResponse:

  def __init__(self, **kwargs):
    self.event: str = kwargs.get('event')
    self.trace_id: str = kwargs.get('trace_id')
    self.status: int = kwargs.get('status')
    self.time: str = kwargs.get('time') if kwargs.get('time') else date_utils.get_date_time()
    self.msg: str = kwargs.get('msg')
    self.body = kwargs.get('body')

  def __str__(self):
    res_str = ""
    for k, v in self.__dict__.items():
      res_str += "%s=%s," % (k, v)
    return res_str[:-1]


class MQTTUtils:
  def __init__(self, broker_address, port=1883, publish_topic=None, client_id=None, keepalive=60, clean_session=False,
               userdata=None, protocol=mqtt.MQTTv311, message_handle=None, broker_encrypt_key=None):
    self.publish_topic = publish_topic
    self.broker_address = broker_address
    self.port = port
    self.client_id = client_id if client_id else sys_info.get_machine_code()
    self.keepalive = keepalive
    self.clean_session = clean_session
    self.userdata = userdata
    self.protocol = protocol
    self.message_handle = message_handle
    self.broker_encrypt_key = broker_encrypt_key

    self.client = mqtt.Client(CallbackAPIVersion.VERSION2, client_id=self.client_id, clean_session=self.clean_session,
                              userdata=self.userdata, protocol=self.protocol)
    self.client.on_connect = self._on_connect
    self.client.on_message = self._on_message_thread
    self.client.on_disconnect = self._on_disconnect
    if self.publish_topic:
      will_msg = {
        "event": "offline",
        "client_id": self.client_id,
        "trace_id": string_tools.get_uuid(),
        "time": date_utils.get_date_time(),
        "body": {"type": "will_msg"}
      }
      self.client.will_set(self.publish_topic, cjson.dumps(will_msg), 2, False)
    self.connected = False
    self._publish_trace_id = []
    self._received_message: Dict[str: MQTTResponse] = {}

  def _on_connect(self, client, userdata, flags, rc, properties):
    if rc == 0:
      log.info("mqtt服务连接成功")
      self.connected = True
    else:
      log.info("mqtt服务连接失败, broker地址: %s:%s, rc: %s" % (self.broker_address, self.port, rc))

  def _on_message_thread(self, client, userdata, message):
    thread_pool.submit(self._on_message, client=client, userdata=userdata, message=message)

  def _on_message(self, client, userdata, message):
    # try:
    if message.topic != self.publish_topic and message.payload:
      mqtt_msg = cjson.loads(message.payload.decode())
      trace_id = mqtt_msg.get('trace_id')
      if trace_id in self._publish_trace_id:
        mqtt_resp = MQTTResponse(**mqtt_msg)
        self._received_message[trace_id] = mqtt_resp
        self._publish_trace_id.remove(trace_id)
      else:
        mqtt_req = MQTTRequest(**mqtt_msg)

        try:
          if mqtt_req.event not in NO_LOG_EVENT:
            log.debug("异步接收mqtt消息: %s" % str(mqtt_req))
          if mqtt_req.body:
            try:
              mqtt_req.body = self.decrypt_body(mqtt_req.body)
            except Exception:
              if mqtt_req.event != MQTTEvent.HAND_SHARK:
                log.error('解密消息失败: {}'.format(mqtt_req))
                return
          if isinstance(mqtt_req.body, dict):
            mqtt_req.body = DictToObj(mqtt_req.body)
        except Exception as e:
          log.info("异步接收mqtt消息异常: %s" % e)
        self.message_handle(mqtt_req, message.topic)
    # except Exception as e:
    #   log.error("接收mqtt消息异常: %s" % e)

  def _on_disconnect(self, client, userdata, flags, rc, properties):
    self.connected = False
    log.info("mqtt服务已断开连接, rc: %s" % rc)

  def encrypt_body(self, body):
    if self.broker_encrypt_key:
      try:
        return sm_tools.encrypt_with_sm4(self.broker_encrypt_key, cjson.dumps(body))
      except Exception:
        return sm_tools.encrypt_with_sm4(self.broker_encrypt_key, body)

  def decrypt_body(self, body):
    if self.broker_encrypt_key:
      try:
        return cjson.loads(sm_tools.decrypt_with_sm4(self.broker_encrypt_key, body))
      except Exception:
        sm_tools.decrypt_with_sm4(self.broker_encrypt_key, body)

  def connect(self, username=None, password=None, timeout=10):
    if username and password:
      self.client.username_pw_set(username, password)
    self.client.connect_async(host=self.broker_address, port=self.port, keepalive=self.keepalive)
    self.client.loop_start()

    start_time = time.time()
    while not self.connected and time.time() - start_time < timeout:
      time.sleep(0.1)

    if not self.connected:
      log.info("mqtt服务连接超时, broker地址: %s:%s" % (self.broker_address, self.port))

  def publish_resp(self, mqtt_resp: MQTTResponse, topic=None, encrypt=True, qos=2, retain=False):
    """
    发送mqtt响应信息
    :param mqtt_resp: 响应信息, 使用 mqtt_service.init_resp('ok', req)
    :param topic: 主题, 默认用订阅的主题
    :param encrypt: 是否加密
    :param qos: 消息质量
    :param retain: 是否保留
    """
    topic = topic if topic else self.publish_topic
    if self.connected:
      if mqtt_resp:
        try:
          if encrypt and mqtt_resp.body:
            mqtt_resp.body = self.encrypt_body(mqtt_resp.body)
          mqtt_resp.client_id = self.client_id
          mqtt_data = cjson.dumps(mqtt_resp)
          self.client.publish(topic, mqtt_data, qos=qos, retain=retain)
          log.debug("异步发送mqtt响应消息: topic: %s message: %s" % (topic, mqtt_data))
        except Exception as e:
          log.error('异步发送mqtt响应消息异常: %s' % e)
    else:
      log.info("异步发送mqtt响应-mqtt服务未连接, 无法发送消息")

  def publish_async(self, message, event: MQTTEvent = None, encrypt=True, topic=None, qos=2, retain=False):
    topic = topic if topic else self.publish_topic
    if self.connected:
      try:
        if encrypt and message:
          message = self.encrypt_body(message)

        mqtt_req = MQTTRequest()
        mqtt_req.client_id = self.client_id
        mqtt_req.body = message
        if event:
          mqtt_req.event = event.value
        mqtt_data = cjson.dumps(mqtt_req)
        self.client.publish(topic, mqtt_data, qos=qos, retain=retain)
        if event not in NO_LOG_EVENT:
          log.debug("异步发送mqtt消息: topic: %s message: %s" % (topic, mqtt_data))
      except Exception as e:
        log.error('异步发送mqtt消息异常: %s' % e)
    else:
      log.info("异步发送mqtt消息-mqtt服务未连接, 无法发送消息")

  def publish_sync(self, message, event: MQTTEvent, topic=None, encrypt=True,
                   qos=2, retain=False, timeout: int = 15, send_time: str = None) -> MQTTResponse:
    if not event: raise Exception("事件不能为空")
    mqtt_resp = MQTTResponse()
    mqtt_resp.status = 500
    mqtt_resp.msg = "服务端未响应, 请联系管理员!"
    topic = topic if topic else self.publish_topic
    if self.connected:
      try:
        if encrypt and message:
          message = self.encrypt_body(message)

        mqtt_req = MQTTRequest()
        mqtt_req.source = "client"
        mqtt_req.client_id = self.client_id
        mqtt_req.body = message
        mqtt_req.time = send_time if send_time else mqtt_req.time
        mqtt_req.event = event.value
        mqtt_data = cjson.dumps(mqtt_req)
        self._publish_trace_id.append(mqtt_req.trace_id)
        self.client.publish(topic, mqtt_data, qos=qos, retain=retain)
        if event not in NO_LOG_EVENT:
          log.debug("同步发送mqtt消息: topic: %s message: %s" % (topic, mqtt_data))
        is_timeout = False
        start_time = time.time()
        while True:
          if mqtt_req.trace_id in self._received_message:
            mqtt_resp = self._received_message.pop(mqtt_req.trace_id)
            if mqtt_resp.body and encrypt:
              mqtt_resp.body = self.decrypt_body(mqtt_resp.body)
            break
          if (time.time() - start_time) >= timeout:
            is_timeout = True
            break
          time.sleep(0.5)
        if not is_timeout:
          if event not in NO_LOG_EVENT:
            log.debug("同步接收mqtt消息: topic: %s message: %s" % (topic, mqtt_resp))
        else:
          log.info("同步发送mqtt消息, 等待返回信息超时: %s" % mqtt_req)
      except Exception as e:
        log.error('同步发送mqtt消息异常: %s-%s' % (event.value, e))
    else:
      log.info("同步接收mqtt消息-mqtt服务未连接, 无法发送消息")
    return mqtt_resp

  def subscribe(self, *topic, qos=0):
    if self.connected:
      if topic and len(topic) > 0:
        for t in topic:
          self.client.subscribe(t, qos=qos)
          log.info("已订阅mqtt主题[%s]" % t)
    else:
      log.info("mqtt服务未连接, 无法订阅主题")

  def unsubscribe(self, *topic):
    if self.connected:
      if topic and len(topic) > 0:
        for t in topic:
          self.client.unsubscribe(t)
          log.info("已解除订阅mqtt主题[%s]" % t)
    else:
      log.info("mqtt服务未连接, 无法解除订阅主题")

  def disconnect(self):
    if self.connected:
      self.client.disconnect()
      self.client.loop_stop()
      log.info("已关闭mqtt服务连接")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
__author__ = 'haoyang'
__date__ = '2024/9/5 10:39'

import time
from threading import Thread, Lock

from kafka import KafkaProducer, errors, KafkaConsumer
from kafka.producer.future import FutureRecordMetadata

from ctools import thread_pool
from ctools.cjson import dumps

"""
import time
from datetime import datetime

from ctools import thread_pool, string_tools
from ctools.ckafka import CKafka

c = CKafka(kafka_url='192.168.3.160:9094', secure=True)

producer = c.init_producer()
consumer = c.init_consumer(enable_auto_commit=False)

def send_msg():
  while True:
    command = input('发送消息: Y/n \n')
    if command.strip() not in ['N', 'n']:
      producer.send_msg('jqxx', '{{"jqid": "{}", "xxxx": "{}"}}'.format(string_tools.get_snowflake_id(), datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S')))
    else:
      break

thread_pool.submit(send_msg)

def consumer_callback(msg):
  print(msg)
  return True

consumer.receive_msg('jqxx', callBack=consumer_callback)

while True: time.sleep(1)
"""
class KafkaInstance:
  def __init__(self, producer: KafkaProducer, consumer: KafkaConsumer):
    self.start_consumer = False
    self.quited = False
    self.producer = producer
    self.consumer = consumer
    self.consumer_callback = {"topic_key": []}

  # FutureRecordMetadata 可以添加回调, 来监听是否发送成功
  # r.add_callback(lambda x: print(x))
  # r.get() 可以同步获取结果
  def send_msg(self, topic, msg, key: str=None, partition:int=None) -> FutureRecordMetadata:
    if self.producer is None: raise RuntimeError("Producer is not initialized")
    if self.quited: return
    return self.producer.send(topic=topic, value=msg, key=None if key is None else key.encode('utf-8'), partition=partition)

  def receive_msg(self, topics: str, callBack=print):
    if self.consumer is None: raise RuntimeError("Consumer is not initialized")
    for topic in topics.split(','):
      if topic not in self.consumer_callback.keys():
        self.consumer_callback[topic] = []
        self.consumer.subscribe(self.consumer_callback.keys())
      self.consumer_callback[topic].append(callBack)
    if not self.start_consumer:
      t = Thread(target=self._start_consumer_poll, daemon=True)
      t.start()

  def _start_consumer_poll(self):
    self.start_consumer = True
    for msg in self.consumer:
      if self.quited: break
      funcList = []
      begin_time = time.time()
      for func in self.consumer_callback[msg.topic]:
        if self.quited: break
        res = func(msg)
        if not self.consumer.config['enable_auto_commit'] and res: self.consumer.commit()
        funcList.append(func.__name__)
      end_time = time.time()
      if end_time - begin_time > 1: print(f"kafka consume too slow!!! {funcList} time cost: ", f'{round(end_time - begin_time, 2)}s')
      funcList.clear()

  def shutdown(self):
    self.quited = True
    try: self.consumer.close()
    except Exception: pass
    try: self.producer.close()
    except Exception: pass


class CKafka:

  def __init__(self, kafka_url: str = '127.0.0.1:9092', secure: bool = False, username: str = 'client', password: str = 'hylink_user_password'):
    self.kafka_url = kafka_url
    self.secure = secure
    self.username = username
    self.password = password

  def init_producer(self, acks=1) -> KafkaInstance:
    print("[ Producer ] Connecting to Kafka [{}]".format(self.kafka_url))
    for i in range(0, 6):
      try:
        if self.secure:
          producer = KafkaProducer(
            acks=acks,
            bootstrap_servers=self.kafka_url,
            value_serializer=lambda x: dumps(x).encode('utf-8'),
            sasl_plain_username=self.username,
            sasl_plain_password=self.password,
            security_protocol='SASL_PLAINTEXT',
            sasl_mechanism='PLAIN'
          )
        else:
          producer = KafkaProducer(
            acks=acks,
            bootstrap_servers=self.kafka_url,
            value_serializer=lambda x: dumps(x).encode('utf-8')
          )
        print("[ Producer ] Success Connected to Kafka [{}]".format(self.kafka_url))
        return KafkaInstance(producer=producer, consumer=None)
      except errors.NoBrokersAvailable:
        print("[ Producer ] Waiting for Kafka [{}] to become available...".format(self.kafka_url))
        time.sleep(3)
    raise RuntimeError("[ Producer ] Failed to connect to Kafka [{}] within 60 seconds".format(self.kafka_url))

  def init_consumer(self, client_id: str = 'ck-py-kafka-consumer', consumer_group: str = 'ck-py-kafka-consumer', enable_auto_commit: bool = True) -> KafkaInstance:
    print("[ Consumer ] Connecting to Kafka [{}]".format(self.kafka_url))
    for i in range(0, 6):
      try:
        if self.secure:
          consumer = KafkaConsumer(
            client_id=client_id,
            group_id=consumer_group,
            enable_auto_commit=enable_auto_commit,
            bootstrap_servers=self.kafka_url,
            value_deserializer=lambda x: x.decode('utf-8'),
            sasl_plain_username=self.username,
            sasl_plain_password=self.password,
            security_protocol='SASL_PLAINTEXT',
            sasl_mechanism='PLAIN'
          )
        else:
          consumer = KafkaProducer(
            client_id=client_id,
            group_id=consumer_group,
            enable_auto_commit=enable_auto_commit,
            bootstrap_servers=self.kafka_url,
            value_deserializer=lambda x: x.decode('utf-8')
          )
        print("[ Consumer ] Success Connected to Kafka [{}]".format(self.kafka_url))
        return KafkaInstance(producer=None, consumer=consumer)
      except errors.NoBrokersAvailable:
        print("[ Consumer ] Waiting for Kafka [{}] to become available...".format(self.kafka_url))
        time.sleep(3)
    raise RuntimeError("[ Consumer ] Failed to connect to Kafka [{}] within 60 seconds".format(self.kafka_url))



import os

from setuptools import setup, find_packages

def parse_requirements(filename):
  with open(filename, 'r') as f:
    lines = f.read().splitlines()
    return [line for line in lines if line and not line.startswith('#')]

requirements = parse_requirements(os.path.join(os.path.dirname(__file__), 'requirements.txt'))

setup(
  name="gomyck-tools",
  version="1.2.10",
  author="gomyck",
  author_email="hao474798383@163.com",
  description="A tools collection for python development by hao474798383",
  long_description=open("README.md").read(),
  long_description_content_type="text/markdown",
  url="https://blog.gomyck.com",
  packages=["ctools"],  # 自动发现并包含包内所有模块
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires=">=3.9",
  install_requires=requirements
)

# 安装依赖, 并生成构建包
# pip install setuptools wheel twine
# rm -rf gomyck_tools.egg-info dist build && python setup.py sdist bdist_wheel && twine upload dist/* && rm -rf gomyck_tools.egg-info dist build

